// CPUID.cpp : Defines the entry point for the console application.
//


// Hans Dorn, Oct-05-2006
#ifdef WINDOWS
#include "stdafx.h"
#include <windows.h>
#endif

#include <stdio.h>

// Hans Dorn, Oct-05-2006
#ifdef WINDOWS
#include <conio.h>
#else
#include <string.h>
#endif

#include "cpuid_x86.h"


float memPerSec(double cycles, int speed)
	{
	if(speed)
		return cycles / speed / 1024 * 100;
	else
		return 33;
	}


#define CPU_REG_KEY		HKEY_LOCAL_MACHINE
#define CPU_REG_SUBKEY	"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0"

#define CPU_SPEED			"~MHz"
#define CPU_IDENTIFIER		"Identifier"
#define CPU_NAME			"ProcessorNameString"
#define CPU_VENDOR			"VendorIdentifier"

//
//
//
void set_registry_key( char *vendor, char *cpu_string )
	{
#ifdef WINDOWS
	HKEY hKey;
	DWORD dwType;
	DWORD dwSize;

	LONG status = RegOpenKeyEx(CPU_REG_KEY, CPU_REG_SUBKEY, 
				 0,
				 KEY_SET_VALUE | KEY_QUERY_VALUE,
				 &hKey);
	if(status != ERROR_SUCCESS)	return;

	dwSize = 1 + strlen( vendor );  // VERY IMPORTANT - ADD + 1
	status = RegSetValueEx(hKey, CPU_VENDOR, NULL, REG_SZ, (LPBYTE)vendor, dwSize);

	dwSize = 1 + strlen( cpu_string );  // VERY IMPORTANT - ADD + 1
	status = RegSetValueEx(hKey, CPU_NAME, NULL, REG_SZ, (LPBYTE)cpu_string, dwSize);

	RegCloseKey(hKey);
#endif
	}

//
//
//
#ifdef WINDOWS
int _tmain(int argc, _TCHAR* argv[])
#else
int main(int argc, char* argv[])
#endif
	{	
	CPU_INFO	our_cpu;

	if(argc > 1) {
		// Does the user want "terse" output?
		if( strcmp( argv[1], "-terse") == 0) {
			printf( " CPUID: '%s %s %s (%s)'\n ID: %X-%X-%X/$%02X%x_  Model:$%02X/$%03X\n"
				, our_cpu.vendor_name()
				, our_cpu.generation()
				, our_cpu.model()
				, our_cpu.core()
				, our_cpu.s_family(), our_cpu.s_model(), our_cpu.s_rev()
				, our_cpu.x_family(), our_cpu.x_model()
				, our_cpu.cpuid_(1,1) & 0xFF
				, our_cpu.x_brand()
				);

			int	dies = our_cpu.cpu_cnt();
			int cores = our_cpu.core_cnt();
			int threads = our_cpu.thread_cnt();
			printf( "dies: %d cores: %d threads: %d  cache: L1=%dK  L2=%dK  L3=%d\n"
				, dies, cores, threads
				, our_cpu.cacheL1()/1024
				, our_cpu.cacheL2()/1024
				, our_cpu.cacheL3()/1024
				);

			printf( " features: %s%s%s%s%s%s%s \n\n"
				, our_cpu.mmx() ? "mmx " : ""
				, our_cpu._3Dnow() ? "3Dnow " : ""
				, our_cpu._3DnowPlus() ? "3Dnow+ " : ""
				, our_cpu.sse()  ? "sse " : ""
				, our_cpu.sse2() ? "sse2 " : ""
				, our_cpu.sse3() ? "sse3 " : ""
				, our_cpu.ssse3() ? "ssse3 " : ""
				);
			}

		// Does the user want to change the systemr registry to Computed CPU name?
		if( strcmp( argv[1], "-setreg") == 0) 
			{
			char buffer[60], counter[10];
			while( our_cpu.cpu_mhz() == 0)
			;
			int	cpu_count = our_cpu.cpu_cnt() * our_cpu.core_cnt() * our_cpu.thread_cnt();
			counter[0] = 0;
			if( cpu_count > 1) sprintf(counter, "  [x%d]", cpu_count);

			sprintf(buffer, "%s %s (%s) @ %dMhz%s"
				, our_cpu.generation()
				, our_cpu.model()
				, our_cpu.core()
				, our_cpu.cpu_mhz(), counter
				);
			set_registry_key( our_cpu.vendor_name(), buffer );
			}
		return 0;
		}


	//
	// Normal display - no command line parameters & not in special folder
	//  
	printf( " CPUID\n -tag: '%s'\n -fullname: '%s'\nComputed name: '%s %s %s (%s)'\n -notes: %s\n"
		, our_cpu.vendor_id()
		, our_cpu.cpuid_name()
		, our_cpu.vendor_name()
		, our_cpu.generation()
		, our_cpu.model()
		, our_cpu.core()
		, our_cpu.notes()
		);

	int	dies = our_cpu.cpu_cnt();
	int cores = our_cpu.core_cnt();
	int threads = our_cpu.thread_cnt();

	printf( " cpus cores hyperT =threads L1    L2    L3   code\n");
	printf( " %2d *  %2d  *  %2d   = %2d   %4dK %4dK %4dK %4dK\n\n"
		, dies, cores, threads, dies*cores*threads
		, our_cpu.cacheL1()/1024
		, our_cpu.cacheL2()/1024
		, our_cpu.cacheL3()/1024
		, our_cpu.codeL1()/1024
		);
	
	printf( " family-model  Brand  Extd:family-model  Brand [9-30b]\n" 
			"    %X-%X-%X       %02X            %02X-%X       %03X\n\n"
		, our_cpu.s_family()
		, our_cpu.s_model()
		, our_cpu.s_rev()
		, our_cpu.brand()
		, our_cpu.x_family()
		, our_cpu.x_model()
		, our_cpu.x_brand()
		);

	printf( " features: %s%s%s%s%s%s%s%s \n\n"
		, our_cpu.mmx() ? "mmx " : ""
		, our_cpu._3Dnow() ? "3Dnow " : ""
		, our_cpu._3DnowPlus() ? "3Dnow+ " : ""
		, our_cpu.sse()  ? "sse " : ""
		, our_cpu.sse2() ? "sse2 " : ""
		, our_cpu.sse3() ? "sse3 " : ""
		, our_cpu.ssse3() ? "ssse3 " : ""
		, our_cpu.could64() ? "em64t " : ""
		);

	printf( " cpuid   -eax-    -ebx-    -ecx-    -edx-  \n");
	int maxFunc =  our_cpu.cpuid_(0, 0) & 0xFF;
	for(int i=0; i<=4 && i <= maxFunc; i++)
        printf( " 0..%d   %08X %08X %08X %08X\n"
			, i
			, our_cpu.cpuid_(i, 0)
			, our_cpu.cpuid_(i, 1)
			, our_cpu.cpuid_(i, 2)
			, our_cpu.cpuid_(i, 3)
			);
	maxFunc =  our_cpu.testxBit(0, 0, 31) ? our_cpu.xcpuid_(0, 0) & 0xFF : 0;
	if( maxFunc == 0)
		printf(" 8..0   [CPU doesn't have xtd ID]\n");
	else
	for(int i=0; i<=6 && i <= maxFunc; i++)
        printf( " 8..%d   %08X %08X %08X %08X\n"
			, i
			, our_cpu.xcpuid_(i, 0)
			, our_cpu.xcpuid_(i, 1)
			, our_cpu.xcpuid_(i, 2)
			, our_cpu.xcpuid_(i, 3)
			);

	our_cpu.L1_speed();
	while( our_cpu.cpu_mhz() == 0)
		;
	printf("\n < Press [enter] for memory speed >");
#ifdef WINDOWS
	FlushConsoleInputBuffer( stdin );
#endif
	getchar();
	printf("\r Speed=%4d Mhz\n"
		, our_cpu.cpu_mhz());

	CACHE_INFO cache = our_cpu.getCache();
	double cycles = our_cpu.cpu_mhz() * 1e6;

	printf("%10s: %6s %4s %4s\n", "Test  ", "L1 ", "L2 ", "RAM");

	int i;
	for(i = 0; i<3; i++) {
		const char *str, *wstr;
		switch( i ) {
			case 0: wstr = str = "8 ints"; break;
			case 1: wstr = str = our_cpu.sse()	? "4 sse" 
						: our_cpu.mmx() ? "8 mmx" : "no simd"; break;
			case 2: str = "pre-cache";  wstr = "not-cached"; break;
			}
		printf("%10s: %6.0f %5.0f %4.0f - Reads\n", str
			,memPerSec( cycles, cache.L1.read_speed[i] )
			,memPerSec( cycles, cache.L2.read_speed[i] )
			,memPerSec( cycles, cache.ram.read_speed[i] ) );
		printf("%10s: %6.0f %5.0f %4.0f - Writes\n", wstr
			,memPerSec( cycles, cache.L1.write_speed[i] )
			,memPerSec( cycles, cache.L2.write_speed[i] )
			,memPerSec( cycles, cache.ram.write_speed[i] ) );
		}

	printf("\n < Press [enter] key to close >");
#ifdef WINDOWS
	FlushConsoleInputBuffer( stdin );
#endif
	getchar();
#ifdef WINDOWS
	exit(0);
#endif

	return 0;
}

