/* sse2_opcodes.hpp: test bench code for seti@home client 
 Originator: Ben Herndon
 */
#ifndef SSE2_OPCODES_H
    #define SSE2_OPCODES_H

#include "opcodes_SSE.hpp"
#include <emmintrin.h>


// MS Visual C++ Compiler, or Intel, or GCC
#if defined( _MSC_VER ) || defined( __ICC ) || (defined( __GNUC__ ) && defined( __SSE2__ ))

    typedef __m128d VEC_D;
    typedef __m128i VEC_I;

    // load
    #define d_get                    _mm_loadu_pd
    #define d_getA                   _mm_load_pd
    #define d_put                    _mm_storeu_pd
    #define d_putA                   _mm_store_pd

    // math
    #define d_add                    _mm_add_pd
    #define d_sub                    _mm_sub_pd
    #define d_mult                   _mm_mul_pd
    #define d_div                    _mm_div_pd
    #define d_max                    _mm_max_pd
    #define d_xor                    _mm_xor_pd
    #define d_and                    _mm_and_pd

    // shuffle    //_MM_SHUFFLE2(x, y)
    #define d_shuffle                _mm_shuffle_pd
    #define d_copyH2L                _mm_movehl_pd
    #define d_copyL2H                _mm_movelh_pd
    #define d_swapRI(_src)           d_shuffle(_src, _src, _MM_SHUFFLE(0,1) )
    #define d_dupeHI                 _mm_unpackhi_pd        
    #define d_dupeLO                 _mm_unpacklo_pd        

    // conversion
    #define s_extract_32bits(aaaa)   _mm_cvtsi128_si32((VEC_I) aaaa)
    #define d_cvt2s                  _mm_cvtpd_ps

    // cache control
    #define s_put1_NC(ptr, aaaa)     _mm_stream_si32(ptr, s_extract_32bits(aaaa) );
    #define s_fence_reads            _mm_lfence
    #define s_fence_all              _mm_mfence

#if defined( _MSC_VER ) || defined( __ICC )
    #define s_flush                  _mm_clflush
#else
    inline void s_flush(const void *p){    // Hans Dorn
         __asm ( "clflush (%%eax)" : : "a" (p) );
    }
#endif


    #endif // defined( _MSC_VER ) || defined( __ICC ) || (defined( __GNUC__ ) && defined( __SSE2__ ))
#endif // #ifndef SSE2_OPCODES_H
