/***************
 *
 *	opt_os_interface.cpp:	
 *
 * Talk to whatever OS is running to get "things" done
 *	Originator:		Ben Herndon
 *
 *   to be used with seti @ home client code
 *
 */
#include "my_types.h"
#include "opt_os_interface.h"

#if defined( _WIN32 ) || defined( _WIN64 )

#include <windows.h>
#ifdef __cplusplus
	extern "C" {
		#include <ntstatus.h>
		#include <PowrProf.h>
		}
#else
#include <PowrProf.h>
#endif
#include <signal.h>
#include <stdio.h>


DWORD	ourPriority=0;
//DWORD	ourAffinity=0;
//DWORD	sysAffinity=0;
//DWORD	soloAffinity=0;

// This was omitted (according to msdn) from header file
typedef struct _PROCESSOR_POWER_INFORMATION {
	ULONG Number;
	ULONG MaxMhz;
	ULONG CurrentMhz;
	ULONG MhzLimit;
	ULONG MaxIdleState;
	ULONG CurrentIdleState;
	} PROCESSOR_POWER_INFORMATION,
	*PPROCESSOR_POWER_INFORMATION;

#ifndef __LINUX__
typedef ULONG NTSTATUS;
#endif

//
// os_cpu_throttle -  Return O/S CPU throttling percentage
//  return: 100 = Full speed
//
float os_cpu_throttle(void ) 
	{
/*	This code all works properly...however
	it forces the cpuid.exe program to test (before running) for the existence
	of the powerxxx.dll file somehwere in the path or windows directory
	and this doesn't exist on Windows NT, or Win 98 - making the program not work
	so we commented it out.
	It usually only shows non 100% on laptops running low on battery anyway.

   PPROCESSOR_POWER_INFORMATION PowerInfo;
	
	ULONG psize = sizeof( *PowerInfo ) * os_cpu_count();
    PowerInfo = (PPROCESSOR_POWER_INFORMATION) calloc( psize, 1 );

	NTSTATUS status;
	status = CallNtPowerInformation(ProcessorInformation, NULL, 0, PowerInfo, psize );
    if( status == STATUS_SUCCESS  )
		{
		return float(PowerInfo[0].CurrentMhz) / PowerInfo[0].MaxMhz * 100.0;
		}
*/
    return 100.00;
	}

//
// os_cpu_count -  Return O/S count of total CPU threads
//
int os_cpu_count(void ) 
	{
	DWORD program, sysAffin;

    SYSTEM_INFO SystemInfo;
    memset( &SystemInfo, NULL, sizeof( SystemInfo ) );
    GetSystemInfo( &SystemInfo );
	int count = SystemInfo.dwNumberOfProcessors;
	if( count == 0) {	
		// That os method didn't work...so now we count up maximum # of affinities
		HANDLE ourProcess = GetCurrentProcess();
		if( GetProcessAffinityMask( ourProcess, &program, &sysAffin ) )
			for(int bit=1; bit ; bit << 1)
				if( sysAffin & bit) count++;
		}

	return count ? count : 1;
	}
/*  
//
// single_cpu -  Run on one processor only
//
void single_cpu(int which)
	{
	HANDLE ourProcess = GetCurrentProcess();

	if( GetProcessAffinityMask( ourProcess, &ourAffinity, &sysAffinity ) ) 
		{
		// Choose affinity bit for 'which' thread
		for(DWORD bit=1<<which; bit ; bit >>= 1 )
			if( sysAffinity & bit)	break;

		if(bit) {
			soloAffinity = bit;
			SetProcessAffinityMask( ourProcess, soloAffinity );
			// Some older O/Ss might not change affinity til a context switch
			// ..so force a switch
			Sleep(0);
			}
		}
	}
*/
//
// ramming_speed -  Setup current process for Maximum cpu usage and only one processor thread
//
void ramming_speed(void)
	{
	HANDLE ourProcess = GetCurrentProcess();

//  single_cpu(16);	// Arbitrarily high cpu #

	ourPriority = GetPriorityClass( ourProcess );
	if( ourPriority )
		if( !SetPriorityClass( ourProcess, REALTIME_PRIORITY_CLASS) )
			fprintf(stderr, "priority miss\\n");
	}

//
// cruising_speed - Reduce priority and thread use of this proces to its normal
//
void cruising_speed(void)
	{
	HANDLE ourProcess = GetCurrentProcess();

	if( !ourPriority)
		ourPriority = GetPriorityClass( ourProcess );
	else
		SetPriorityClass( ourProcess, ourPriority);
//	if( ourAffinity )
//		SetProcessAffinityMask( ourProcess, ourAffinity );
	}

//
// hires_timer - Highest resolution timer available in O/S
//
uint64 hires_timer(uint32 &resolution)
	{
	const int ten_million = 10000000;
    FILETIME sysTime;
    LARGE_INTEGER time;
    GetSystemTimeAsFileTime(&sysTime);
    time.LowPart  = sysTime.dwLowDateTime;
    time.HighPart = sysTime.dwHighDateTime;  // Time is in 100 ns units
	resolution = ten_million;
	return time.QuadPart;
	}

// HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Memory Management\\SecondLevelDataCache

#define CPU_REG_KEY		HKEY_LOCAL_MACHINE
#define CPU_REG_SUBKEY	"SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Memory Management\\"

#define L2_DATA_CACHE		"SecondLevelDataCache"


//
// os_L2_cache_size - Get value from registry: Might be the off-cpu cache size
//
int os_L2_cache_size( void )
	{
	HKEY hKey;
	DWORD dwType;
	DWORD dwSize;
	DWORD L2_size = 0;

	LONG status = RegOpenKeyEx(CPU_REG_KEY, CPU_REG_SUBKEY, 
				 0,
				 KEY_QUERY_VALUE,
				 &hKey);
	if(status != ERROR_SUCCESS)	return 0;

	dwSize = sizeof( L2_size );
	status = RegQueryValueEx(hKey, L2_DATA_CACHE, NULL, NULL, (LPBYTE)&L2_size, &dwSize);
	RegCloseKey(hKey);

	return L2_size;
	}


//*************************************************************************
#else  // Not windows
//*************************************************************************

#include <unistd.h>		// Hans Dorn, Oct-05-2006
#include <stdio.h>
#include <fstream>
#include <sched.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>


//
// os_cpu_throttle -  Return O/S CPU throttling percentage
//  return: 100 = Full speed
//
float os_cpu_throttle(void ) 
	{
    return 100.00;
	}

//
// os_cpu_count - various unixes
//
int os_cpu_count(void ) {
	int	cpu_count = 1;
	// sysconf not working on OS2
#if defined(_SC_NPROCESSORS_ONLN) && !defined(__EMX__)
		cpu_count = sysconf(_SC_NPROCESSORS_ONLN);
	#elif defined(HAVE_SYS_SYSCTL_H) && defined(CTL_HW) && defined(HW_NCPU)
		// Get number of CPUs
		mib[0] = CTL_HW;
		mib[1] = HW_NCPU;
		int len = sizeof(cpu_count);
		sysctl(mib, 2, &cpu_count, &len, NULL, 0);
	#elif defined(_HPUX_SOURCE)
		struct pst_dynamic psd; 
		pstat_getdynamic ( &psd, sizeof ( psd ), (size_t)1, 0 );
		cpu_count = psd.psd_proc_cnt;
	#else
	#error Need to specify a sysconf() define to obtain number of processors
	#endif
	return cpu_count;
	}


int			ourPriority;
cpu_set_t	ourAffinity;
cpu_set_t	sysAffinity;
cpu_set_t	soloAffinity;


/* Declared in "sched.h" !!!   Hans Dorn, Oct-05-2006
       void CPU_CLR(int cpu, cpu_set_t *set);
       int CPU_ISSET(int cpu, cpu_set_t *set);
       void CPU_SET(int cpu, cpu_set_t *set);
       void CPU_ZERO(cpu_set_t *set);
       int getpriority(int which, int who);
       int setpriority(int which, int who, int prio);
*/


//
// ramming_speed -  Setup current process for Maximum cpu usage and only one processor thread
//
void ramming_speed(void)
	{
    if( sched_getaffinity(0 /* us */,  sizeof(cpu_set_t),  &ourAffinity) == 0)
		{
			// Choose maximum hyperthread or core
		int cpu;
		for(cpu=16; cpu ; cpu-- )
			if( CPU_ISSET( cpu, &ourAffinity))	break;
		single_cpu( cpu );
		}
	ourPriority = getpriority(PRIO_PROCESS, 0 /* us */);
	if(ourPriority > -10 ) 	// priority by Hans Dorn
		{
		setpriority(PRIO_PROCESS, 0 /* us */, -10);
		}

	}

//
// cruising_speed - Reduce priority and thread use of this proces to its normal
//
void cruising_speed(void)
	{
    sched_setaffinity(0 /* us */,  sizeof(cpu_set_t),  &ourAffinity);
    setpriority(PRIO_PROCESS, 0 /* us */, ourPriority);	// Hans Dorn
	}

//
// single_cpu -  Run on one processor only
//
void single_cpu(int which)
	{
	CPU_ZERO(&soloAffinity);
	CPU_SET(which, &soloAffinity);
	sched_setaffinity(0 /* us */,  sizeof(cpu_set_t),  &soloAffinity);
	}


//
// hires_timer - Highest resolution timer available in O/S
//  unix ver by Hans Dorn
uint64 hires_timer(uint32 &resolution)
{
    uint64 result;
    struct timeval tv; 
    struct timezone tz;

    gettimeofday(&tv, &tz);

//  resolution = 1000000;
    resolution = 1000000;
    result = tv.tv_sec;
    result *= (uint64) 1000000L;
    result += (uint64) tv.tv_usec;
    return result;
}


//
// unix_proc_info - Search "/proc" file for specified tag line 
//                - return value past ':'
//
char *unix_proc_info( char *search_tag )
	{
    char buf[256];
    static char value[40];
    
    value[0] = 0;
    FILE* f = fopen("/proc/cpuinfo", "r");
    if (!f) return NULL;

    while (fgets(buf, 256, f)) {
        if ( strstr(buf, search_tag) == buf ) {
        	char *p = strstr(buf, ":") + 1;
        	while(*p == ' ')
        		p++;
            strcpy(value, p);
            p = strchr(value, '\\n');
            if (p) { *p = '\\0'; }
        	}
		}
    fclose(f);
    return value;
	}

	
//
// os_L2_cache_size - Get value from /proc : Might be the off-cpu cache size
//
int os_L2_cache_size( void )
	{
	char *value;
	
	if( (value = unix_proc_info( "cache size" )) )
		return atoi( value );
	if( (value = unix_proc_info( "L2 cache" )) )
		return atoi( value );
	if( (value = unix_proc_info( "D-cache" )) )	// There is also an "I-cache"
		return atoi( value );
	
	return 0;
	}



#endif



uint64 hires_edge(void)
	{
	uint64	now, then;
	uint32		rez;

	now = hires_timer( rez );
	do	{	 // kill time until it changes
		then = hires_timer( rez );
		} while( now == then );
	return then;
	}
