/*************************************
 *
 * SSE folding subroutines from SSE2 by Alex Kan
 *
 * Modified by Joe Segur
 *
 * Must be compiled for SSE!
 */

// In order to use IPP, set -DUSE_IPP and one of -DUSE_SSE3, -DUSE_SSE2,
// -DUSE_SSE or nothing(generic), IPP precedes FFTW, ooura 
// TMR
/*
Abbreviation    Meaning

px  Generic C optimized code
a6  Intel Pentium III processor code with SSE optimization
w7  Intel Pentium 4 processor code with SSE2 optimization
t7  Intel Pentium 4 processor code with SSE3 optimization
m7  Intel Pentium D processors
m7  Intel Xeon processor with Intel EM64T optimization
i7  Intel Itanium processor and Itanium 2 processor optimization
v8	Intel Core 2-based processors with SSSE3 optimization
s2  Intel IXP4XX Product Line of Network Processor optimization
*/
#if defined(USE_IPP)
#pragma message ("-----IPP-----")
#if defined(USE_SSSE3)				  
#define V8 1						  
#pragma message ("-----SSSE3/tm-----")
#include <ipp_v8.h>					  
#include <tmmintrin.h> // SSSE3       
#include <fvec.h>                     
#elif defined(USE_SSE3)				  
#define T7 1
#pragma message ("-----SSE3/pm-----") 
#include <ipp_t7.h>
#include <pmmintrin.h> // SSE3        
#include <fvec.h>                     
#elif defined(USE_SSE2)
#define W7 1
#pragma message ("-----SSE2/em-----")
#include <emmintrin.h> //SSE2        
#include <fvec.h>                    
#include <ipp_w7.h>
#elif defined(USE_SSE)
#define A6 1
#pragma message ("-----SSE/xm-----") 
#include <xmmintrin.h> //SSE         
#include <fvec.h>                    
#include <ipp_a6.h>
#else
#pragma message ("-----MMX-----")
#include <ipp_px.h>
#endif
#include <ipp.h>
#include <ippvm.h>       
#include <ipps.h>               
#endif

//#include "stdafx.h"
#include "opt_SSE.hpp"

#include "../pulsefind.h"
#include "../seti.h"

// #define ZERO  ((__m128) {0.0f, 0.0f, 0.0f, 0.0f})

typedef union submask {  float    f[4];  __m128   v;  };
submask tailmask[4] = {
  { 0.0f,  0.0f,   0.0f,   0.0f},
  { 0.0f, 1e37f,  1e37f,  1e37f},
  { 0.0f,  0.0f,  1e37f,  1e37f},
  { 0.0f,  0.0f,   0.0f,  1e37f}
};


sum_func_ptt(foldArrayBy3) {
  float max;
  __m128 maxV = ZERO;
  int i = 0;
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1;
  float *pst = P->dest;
  __m128 x1, x2, x3, xs12, xs;
  
  while (i < P->di-4) {
    x1 = _mm_load_ps(p1+i);
    x2 = _mm_loadu_ps(p2+i);
    x3 = _mm_loadu_ps(p3+i);
    xs12 = _mm_add_ps(x1, x2);
    xs = _mm_add_ps(xs12, x3);
    _mm_store_ps(pst+i, xs);
    maxV = _mm_max_ps(maxV, xs);
    i += 4;
  }
  x1 = _mm_load_ps(p1+i);
  x2 = _mm_loadu_ps(p2+i);
  x3 = _mm_loadu_ps(p3+i);
  xs12 = _mm_add_ps(x1, x2);
  xs = _mm_add_ps(xs12, x3);
  
  _mm_store_ps(pst+i, xs);
  xs = _mm_sub_ps(xs, tailmask[P->di & 0x3].v);
  maxV = _mm_max_ps(maxV, xs);
  
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x4e));
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x39));
  _mm_store_ss(&max, maxV);
  
  return max;
}


sum_func_ptt(foldArrayBy3LO) {
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1;
  float *pst = P->dest;
  float max = 0.0f;
  int i = 0;
  
  while (i < P->di) {
    pst[i] = p1[i] + p2[i] + p3[i];
    if (pst[i] > max) max = pst[i];
    i += 1;
  }
  return max;
}
sum_func AKSSETB3[FOLDTBLEN] = {
    foldArrayBy3LO, foldArrayBy3LO, foldArrayBy3LO, foldArrayBy3LO,
    foldArrayBy3LO, foldArrayBy3LO, foldArrayBy3LO, foldArrayBy3LO,
    foldArrayBy3,   foldArrayBy3LO, foldArrayBy3LO, foldArrayBy3LO,
    foldArrayBy3
  };

sum_func_ptt(foldArrayBy4) {
  float max;
  __m128 maxV = ZERO;
  int i = 0;
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1, *p4 = ss[0]+P->tmp2;
  float *pst = P->dest;
  __m128 x1, x2, x3, x4, xs12, xs34, xs;
  
  while (i < P->di-4) {
    x1 = _mm_load_ps(p1+i);
    x2 = _mm_loadu_ps(p2+i);
    x3 = _mm_loadu_ps(p3+i);
    x4 = _mm_loadu_ps(p4+i);
    xs12 = _mm_add_ps(x1, x2);
    xs34 = _mm_add_ps(x3, x4);
    xs = _mm_add_ps(xs12, xs34);
    _mm_store_ps(pst+i, xs);
    maxV = _mm_max_ps(maxV, xs);
    i += 4;
  }
  x1 = _mm_load_ps(p1+i);
  x2 = _mm_loadu_ps(p2+i);
  x3 = _mm_loadu_ps(p3+i);
  x4 = _mm_loadu_ps(p4+i);
  xs12 = _mm_add_ps(x1, x2);
  xs34 = _mm_add_ps(x3, x4);
  xs = _mm_add_ps(xs12, xs34);
  
  _mm_store_ps(pst+i, xs);
  xs = _mm_sub_ps(xs, tailmask[P->di & 0x3].v);
  maxV = _mm_max_ps(maxV, xs);
  
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x4e));
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x39));
  _mm_store_ss(&max, maxV);
  
  return max;
}

sum_func_ptt(foldArrayBy4LO) {
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1, *p4 = ss[0]+P->tmp2;
  float *pst = P->dest;
  float max = 0.0f;
  int i = 0;
  
  while (i < P->di) {
    pst[i] = p1[i] + p2[i] + p3[i] + p4[i];
    if (pst[i] > max) max = pst[i];
    i += 1;
  }
  return max;
}
sum_func AKSSETB4[FOLDTBLEN] = {
    foldArrayBy4LO, foldArrayBy4LO, foldArrayBy4LO, foldArrayBy4LO,
    foldArrayBy4LO, foldArrayBy4LO, foldArrayBy4LO, foldArrayBy4LO,
    foldArrayBy4,   foldArrayBy4LO, foldArrayBy4LO, foldArrayBy4LO,
    foldArrayBy4
  };

sum_func_ptt(foldArrayBy5) {
  float max;
  __m128 maxV = ZERO;
  int i = 0;
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1, *p4 = ss[0]+P->tmp2, *p5 = ss[0]+P->tmp3;
  float *pst = P->dest;
  __m128 x1, x2, x3, x4, x5, xs12, xs34, xs125, xs;
  
  while (i < P->di-4) {
    x1 = _mm_load_ps(p1+i);
    x2 = _mm_loadu_ps(p2+i);
    x3 = _mm_loadu_ps(p3+i);
    x4 = _mm_loadu_ps(p4+i);
    x5 = _mm_loadu_ps(p5+i);
    xs12 = _mm_add_ps(x1, x2);
    xs34 = _mm_add_ps(x3, x4);
    xs125 = _mm_add_ps(xs12, x5);
    xs = _mm_add_ps(xs34, xs125);
    _mm_store_ps(pst+i, xs);
    maxV = _mm_max_ps(maxV, xs);
    i += 4;
  }
  x1 = _mm_load_ps(p1+i);
  x2 = _mm_loadu_ps(p2+i);
  x3 = _mm_loadu_ps(p3+i);
  x4 = _mm_loadu_ps(p4+i);
  x5 = _mm_loadu_ps(p5+i);
  xs12 = _mm_add_ps(x1, x2);
  xs34 = _mm_add_ps(x3, x4);
  xs125 = _mm_add_ps(xs12, x5);
  xs = _mm_add_ps(xs34, xs125);
  
  _mm_store_ps(pst+i, xs);
  xs = _mm_sub_ps(xs, tailmask[P->di & 0x3].v);
  maxV = _mm_max_ps(maxV, xs);
  
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x4e));
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x39));
  _mm_store_ss(&max, maxV);
  
  return max;
}

sum_func_ptt(foldArrayBy5LO) {
  
  const float *p1 = ss[0], *p2 = ss[0]+P->tmp0, *p3 = ss[0]+P->tmp1, *p4 = ss[0]+P->tmp2, *p5 = ss[0]+P->tmp3;
  float *pst = P->dest;
  float max = 0.0f;
  int i = 0;
  
  while (i < P->di) {
    pst[i] = p1[i] + p2[i] + p3[i] + p4[i] + p5[i];
    if (pst[i] > max) max = pst[i];
    i += 1;
  }
  return max;
}

sum_func AKSSETB5[FOLDTBLEN] = {
    foldArrayBy5LO, foldArrayBy5LO, foldArrayBy5LO, foldArrayBy5LO,
    foldArrayBy5LO, foldArrayBy5LO, foldArrayBy5LO, foldArrayBy5LO,
    foldArrayBy5,   foldArrayBy5LO, foldArrayBy5LO, foldArrayBy5LO,
    foldArrayBy5
  };

// 2A version allows non-aligned tmp0

sum_func_ptt(foldArrayBy2A) {
  float max;
  __m128 maxV = ZERO;
  int i = 0;
  
  const float *p1 = ss[1]+P->offset, *p2 = ss[1]+P->tmp0;
  float *pst = P->dest;
  __m128 x1, x2, xs;
  
  while (i < P->di-4) {
    x1 = _mm_load_ps(p1+i);
    x2 = _mm_loadu_ps(p2+i);
    xs = _mm_add_ps(x1, x2);
    _mm_store_ps(pst+i, xs);
    maxV = _mm_max_ps(maxV, xs);
    i += 4;
  }
  x1 = _mm_load_ps(p1+i);
  x2 = _mm_loadu_ps(p2+i);
  xs = _mm_add_ps(x1, x2);
  
  _mm_store_ps(pst+i, xs);
  xs = _mm_sub_ps(xs, tailmask[P->di & 0x3].v);
  maxV = _mm_max_ps(maxV, xs);
  
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x4e));
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x39));
  _mm_store_ss(&max, maxV);
  
  return max;
}

// 2AL version requires aligned tmp0


sum_func_ptt(foldArrayBy2AL) {
  float max;
  __m128 maxV = ZERO;
  int i = 0;
  
  const float *p1 = ss[1]+P->offset, *p2 = ss[1]+P->tmp0;
  float *pst = P->dest;
  __m128 x1, x2, xs;
  
  while (i < P->di-4) {
    x1 = _mm_load_ps(p1+i);
    x2 = _mm_load_ps(p2+i);
    xs = _mm_add_ps(x1, x2);
    _mm_store_ps(pst+i, xs);
    maxV = _mm_max_ps(maxV, xs);
    i += 4;
  }
  x1 = _mm_load_ps(p1+i);
  x2 = _mm_load_ps(p2+i);
  xs = _mm_add_ps(x1, x2);
  
  _mm_store_ps(pst+i, xs);
  xs = _mm_sub_ps(xs, tailmask[P->di & 0x3].v);
  maxV = _mm_max_ps(maxV, xs);
  
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x4e));
  maxV = _mm_max_ps(maxV, _mm_shuffle_ps(maxV, maxV, 0x39));
  _mm_store_ss(&max, maxV);
  
  return max;
}

sum_func_ptt(foldArrayBy2LO) {
  
  const float *p1 = ss[1]+P->offset, *p2 = ss[1]+P->tmp0;
  float *pst = P->dest;
  float max = 0.0f;
  int i = 0;
  
  while (i < P->di) {
    pst[i] = p1[i] + p2[i];
    if (pst[i] > max) max = pst[i];
    i += 1;
  }
  return max;
}
// Note: Benchmarks need fully initialized 2fold table
sum_func AKSSETB2[FOLDTBLEN] = {
    foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2A,  foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,
    foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,
    foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,
    foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,
    foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A,  foldArrayBy2A
  };

sum_func AKSSETB2AL[FOLDTBLEN] = {
    foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2AL, foldArrayBy2LO, foldArrayBy2LO, foldArrayBy2LO,
    foldArrayBy2AL
  };

FoldSet AKSSEfold = {AKSSETB3, AKSSETB4, AKSSETB5, AKSSETB2, AKSSETB2AL, "AK SSE"};
