/* opcodes_3dnow.hpp: 

 * Originator: Ben Herndon
 */

#pragma once

#ifndef OPCODES_3DNOW_H
    #define OPCODES_3DNOW_H

#undef s_fill
#undef s_get
#undef s_getA
#undef s_get2
#undef s_put
#undef s_putA
#undef s_get1
#undef s_fetch
#undef s_fetchW
#undef s_putNC
#undef s_add
#undef s_sub
#undef s_mult
#undef s_div
#undef s_and
#undef s_xor
#undef s_max
#undef s_accum
#undef s_swapRI


// AMD 3Dnow inline opcode prototypes -  MS Visual C++ Compiler
//
#if defined( _MSC_VER ) || defined( __ICC )

#include "mm3dnow.h"

typedef __m64   MMX;

#define s_begin()           _m_femms()
#define s_end()             _m_femms()

#define s_get( ptr )        *( (MMX *) (ptr) )
#define s_getA( ptr )       *( (MMX *) (ptr) )
#define s_get2( ptr )       *( (MMX *) (ptr) )
#define s_put( ptr, aa )    ( *((MMX *) (ptr)) = aa )
#define s_putA( ptr, aa )   ( *((MMX *) (ptr)) = aa )
#define s_get1( ptr )       _m_from_float( ptr )
	// Dangerous, this usually moves to mem, then fld and fstp

// #define s_put1( ptr, aa )   ( *ptr = _m_to_float( aa ) )

#define s_fetch( addr)		_m_prefetch( addr )
#define s_fetchW( addr)		_m_prefetchw( addr )
#define s_putNC				_mm_stream_pi
// Math
//#define s_add(aa, bb)	__asm { pfadd  aa, bb }
#define s_add(aa, bb)	_m_pfadd(aa, bb)
#define s_sub(aa, bb)	_m_pfsub(aa, bb)
#define s_mult(aa, bb)	_m_pfmul(aa, bb)
#define s_div(aa, bb)	_m_pfdiv(aa, bb)

#define s_and(aa, bb)	_m_pand(aa, bb)
#define s_xor(aa, bb)	_m_pxor(aa, bb)
#define s_max(aa, bb)	_m_pfmax(aa, bb)
#define s_accum(aa, bb)	_m_pfacc(aa, bb)		// [A+a:B+b] = pfacc(Aa, Bb)

/*
__inline void s_fetch( void *addr) {  _m_prefetch( addr ); }
__inline void s_fetchW( void *addr) {  _m_prefetchw( addr ); }

MMX	s_get( MMX *ptr)			{ return *ptr; }
MMX	s_getA( MMX *ptr)			{ return *ptr; }
MMX	s_get2( MMX *ptr)			{ return *ptr; }
MMX	s_put( MMX *ptr, MMX reg)	{ return *ptr = reg; }
MMX	s_get1( MMX *ptr)			{ return _m_from_float(reg); }
MMX	s_put( float *ptr, MMX reg)	{ return *ptr = _m_to_float( reg ); }


MMX	s_add(MMX one, MMX two)		{ return _m_pfadd( one, two ); }
MMX	s_sub(MMX one, MMX two)		{ return _m_pfsub( one, two ); }
MMX	s_mult(MMX one, MMX two)	{ return _m_pfmul( one, two ); }
MMX s_div(MMX one, MMX two)		{ return _m_pfdiv( one, two ); }
MMX s_xor(MMX one, MMX two)		{ return _m_pxor( one, two ) ; }
MMX s_max(MMX one, MMX two)		{ return _m_pfmax( one, two ); }
MMX s_accum(MMX one, MMX two)	{ return _m_pfacc( one, two ); } // [A+a:B+b] = pfacc(Aa, Bb)
*/

//  s_putNC  - Put to memory direclty, bypass cache

/*
__inline void s_putNC( float *addr, MMX mmreg )
    {
    __asm
        {
        mov eax, addr
        movq mm5, mmreg
        movntq MMWORD PTR[eax], mm5
        }
    }
*/

// 3Dnow PLUS (Athlon+)
        #define s_accumCX( Aa, Bb ) _m_pfpnacc( Aa, Bb )    // [A-a:B+b] = pfpnacc(Aa, Bb)
        #define s_swapRI( aa )      _m_pswapd( aa )

        #define s_fill_I( aa )      ( _m_punpckhdq( aa, aa ) )
        #define s_fill_R( aa )      ( _m_punpckldq( aa, aa ) )
        #define s_fill( init )      ( s_fill_R( s_get1( init ) ) )

    #endif



// GNU Compiler - GCC's versions of 3DNow
#if defined( __GNUC__ ) && defined( __3DNOW__ ) 

#include <mmintrin.h>

typedef float MMX   __attribute__( (mode(V2SF), aligned(16)) );


__inline void s_begin( void )	{ _m_femms(); }
__inline void s_end( void )	    { _m_femms(); }

#define s_get( ptr )        ( *( MMX * ) ptr )
#define s_getA( ptr )       ( *( MMX * ) ptr )
#define s_put( ptr, aa )    ( *(( MMX * ) ptr) = aa )
#define s_putA( ptr, aa )   ( *(( MMX * ) ptr) = aa )
#define s_add( aa, bb )     _m_pfadd( aa, bb )
#define s_sub( aa, bb )     _m_pfsub( aa, bb )
#define s_mult( aa, bb )    _m_pfmul( aa, bb )
#define s_div( aa, bb )     _m_pfdiv( aa, bb )
#define s_xor( aa, bb )     _m_pxor( aa, bb )
#define s_max( aa, bb )     _m_pfmax( aa, bb )

    #endif // defined(__GNUC__)
#endif // #ifndef OPCODES_3DNOW_H
