/* sse_opcodes.hpp: 

 * test bench code for seti@home client 
 * Originator: Ben Herndon
 */
#pragma once


/*    __SSE__        
    UNPCKHPS    (CAca) = (AaBb, CcDd)
    UNPCKLPS    (Dbdb) = (AaBb, CcDd)

    MOVHPS        (MmBb) = (AaBb, Mm)
    MOVLPS        (AaMm) = (AaBb, Mm)
    MOVHLPS        (AaDd) = (AaBb, CcDd)
    MOVLHPS        (CcBb) = (AaBb, CcDd)

    __3DNow__
    +PFNACC        ([b-B][a-A])    = (Aa, Bb)
    +PFPNACC     ([b+B][a-A])    = (Aa, Bb)
    +PSWAPD         (bB)            = (Aa, Bb)
    
    PFACC          ([A+a][B+b])    = (Aa, Bb)
    PFADD          ([A+B][a+b])    = (Aa, Bb)
    PFMUL         ([A*B][a*b])    = (Aa, Bb)
    PFSUB         ([A-B][a-b])    = (Aa, Bb)
    PFSUBR         ([B-A][B-A])    = (Aa, Bb)
    PFMAX         ([max(A,B)][max(a,b)])    = (Aa, Bb)
    PFSUB         ([A-B][a-b])    = (Aa, Bb)
    OPCODE         (xx)    = (Aa, Bb)
*/


// MS Visual C++ Compiler, or Intel, or GCC
#if defined( _MSC_VER ) || defined( __ICC ) || (defined( __GNUC__ ) && defined( __SSE__ ))
    #include <xmmintrin.h>

typedef __m128  XMM;
typedef __m128  VEC;
typedef __m64 *_MP;

    #define s_get( ptr )            _mm_loadu_ps( ptr )

    #define s_getA( ptr )           _mm_load_ps( ptr )

    #define s_put( ptr, aaaa )      _mm_storeu_ps( ptr, aaaa )
    #define s_putA( ptr, aaaa )     _mm_store_ps( ptr, aaaa )
    #define s_add( aaaa, bbbb )     _mm_add_ps( aaaa, bbbb )
    #define s_sub( aaaa, bbbb )     _mm_sub_ps( aaaa, bbbb )
    #define s_mult( aaaa, bbbb )    _mm_mul_ps( aaaa, bbbb )
    #define s_div( aaaa, bbbb )     _mm_div_ps( aaaa, bbbb )
    #define s_xor( aaaa, bbbb )     _mm_xor_ps( aaaa, bbbb )
    #define s_max( aaaa, bbbb )     _mm_max_ps( aaaa, bbbb )
    #define s_and( aaaa, bbbb )     _mm_and_ps( aaaa, bbbb )

#if 1
    #define s_getU( aaaa, ptr )  \
        aaaa = _mm_loadh_pi( _mm_loadl_pi(aaaa, (_MP)ptr), ((_MP)(ptr))+1 )
    #define s_putU( ptr, aaaa ) \
        _mm_storel_pi((_MP)ptr, aaaa), _mm_storeh_pi( ((_MP)ptr)+1 , aaaa)
#else
    #define s_getU( aaaa, ptr )  \
        (aaaa = _mm_loadu_ps( (float *) ptr) )
    #define s_putU( ptr, aaaa ) \
        _mm_storeu_ps( (float *)ptr, aaaa ) 
#endif


    #define s_add1( aaaa, bbbb )            _mm_add_ss( aaaa, bbbb )
    #define s_mult1( aaaa, bbbb )           _mm_mul_ss( aaaa, bbbb )
    #define s_max1( aaaa, bbbb )            _mm_max_ss( aaaa, bbbb )
    #define s_get1( addr )                  _mm_load_ss( addr )
    #define s_put1( addr, bbbb )            _mm_store_ss( addr, bbbb )

    // shuffle
    #define s_shuffle( aaaa, bbbb, mask )   _mm_shuffle_ps( aaaa, bbbb, mask )
    #define s_fill( a )                     _mm_set_ps1( a )
    #define s_copyH2L( one, two )           _mm_movehl_ps( one, two )
    #define s_copyL2H( one, two )           _mm_movelh_ps( one, two )
    #define s_swapRI(_src)                    s_shuffle(_src, _src, _MM_SHUFFLE(2,3,0,1) )


    // Cache management
    #define s_fetch( addr )                 _mm_prefetch( (char*)(addr), _MM_HINT_T0 )
    #define s_fetch_temp( addr )            _mm_prefetch( (char*)(addr), _MM_HINT_NTA )
    #define s_fence_writes                    _mm_sfence
    #define s_putA_nc                        _mm_stream_ps

#endif

