/******************
 *
 * opt_sse.cpp
 * 
 * Description: Intel & AMD SSE optimzized functions
 *  CPUs: Intel Pentium III and beyond,  AMD Athlon model 6 -  and beyond
 *
 */

// Copyright 2004 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "stdafx.h"

#include "opt_SSE.hpp"
#include "chirpfft.h"
#include "pulsefind.h"
#include "mathimf.h"


int	sse_subs[] = {4,1,2,3};
int	sse_mask1[] = {8, 5, 6, 7, 4, 4, 4, 4 };
int	sse_mask2[] = {8, 8, 8, 8, 8, 5, 6, 7 };
ALIGN_SIMD unsigned int sse_partMask[][4] = {
	{ _b1111,_b1111,_b1111,_b1111 }, // 0
	{ _b0000,_b0000,_b0000,_b1111 }, // 1
	{ _b0000,_b0000,_b1111,_b1111 }, // 2
	{ _b0000,_b1111,_b1111,_b1111 }, // 3
	{ _b1111,_b1111,_b1111,_b1111 }, // 4
	{ _b1111,_b0000,_b0000,_b0000 }, // 5
	{ _b1111,_b1111,_b0000,_b0000 }, // 6
	{ _b1111,_b1111,_b1111,_b0000 }, // 7
	{ _b0000,_b0000,_b0000,_b0000 }, // 8
	}; 

ALIGN_SIMD __m128 sse_nums[] = {
	{ 0.0f,0.0f,0.0f,0.0f },
	{ 1.0f,1.0f,1.0f,1.0f },
	{ 2.0f,2.0f,2.0f,2.0f },
	{ 3.0f,3.0f,3.0f,3.0f },

	{1.5707963268, 1.5707963268, 1.5707963268, 1.5707963268},
	{-0.6466386396, -0.6466386396, -0.6466386396, -0.6466386396 },
	{0.0679105987, 0.0679105987, 0.0679105987, 0.0679105987},
	{-0.0011573807, -0.0011573807, -0.0011573807, -0.0011573807},
	{-1.2341299769, -1.2341299769, -1.2341299769, -1.2341299769},
	{0.2465220241, 0.2465220241, 0.2465220241, 0.2465220241},
	{-0.0123926179, -0.0123926179, -0.0123926179, -0.0123926179},
	}; 


ALIGN_SIMD int sse_inums[][4] = {
	{ INT_MIN,	0,	INT_MIN,	0 },
	{ 0,	INT_MIN,	0,	INT_MIN},
	{ 0,		0,		0,		0  },
	};

#define SSE_MASK(idx) (*(VEC *)&sse_partMask[idx])



//=================================================
// sse_f_GetPeak
//=================================================

f_GetPeak_ptt( sse_f_GetPeak )
    {
    float   sum1;
    int     first, last, i;
    int     weight_i;	// f_weight[] is actually middle of a table

	first	= ul_TOffset - ul_HalfSumLength;
    last	= ul_TOffset + ul_HalfSumLength + 1;
    weight_i= first - ul_TOffset;	// negative indexes are ok

	sum1 = 0.0;
	 // Then the going up direction (in the f_weight table)
    for ( i = first; i < last; i++ ) 
		sum1 += ( fp_PoT[i] - f_MeanPower ) * f_weight[weight_i++];

    count_flops( 6 * ul_HalfSumLength );
    return ( sum1 * f_PeakScaleFactor );
    }

//=================================================
// sse_f_sum
//=================================================

// s_addp2f -	
//	function: Given	a 4	packed float input,	return the sum of all 4
//	concept:  Ben Herndon
//
static __forceinline VEC s_addp2s( VEC sum1 )
	{
	VEC		sumReg = sum1;
	sumReg = s_copyH2L(	sumReg,	sum1 );
	sumReg = s_add(	sumReg,	sum1 );		// [0][1] + [3][4]
	sum1 = s_shuffle( sumReg, sumReg, 1 );
	sumReg = s_add1( sumReg, sum1 );	// [1] + [0]

	return ( sumReg );
	}

f_sum_ptt( sse_f_sum2 )
    {
    float   sum = 0;
    for ( int i = length; i; i-- )
        {
        sum += buf[i - 1];
        }
	f_total = sum;
    }

//
//
//
 f_sum_ptt( sse_f_sum )
	{
	VEC		sum1, sum2,	sum3, sum4, tmp1;
	sum1 = sum2 = sum3 = sum4 = ZERO;
	int	i, j, mask;
	const int	stride = 16;
	
	// if mask = 1 - get high 3 and subtract 3
	// if mask = 2 -- get high 2 and subtract 2
	// if mask = 3 -- get high 1 and subtract 1
	// if mask = 0 -- get all or and subtract 4


	i = length;
	if(length > stride-1) {
		mask = ((unsigned int)&buf[i-4])>>2 & 3;
		s_getU(sum4, &buf[i-4] );
		sum4 = s_and(sum4, SSE_MASK( mask ));
		i -= sse_subs[mask];
	
		for	(   ; i > stride - 1; i -= stride )
			{
			s_fetch( &buf[i - 6*stride] );

			sum1 = s_add( sum1,	s_getA(&buf[i-4] ) );
			sum2 = s_add( sum2,	s_getA(&buf[i-8] ) );
			sum3 = s_add( sum3,	s_getA(&buf[i-12] ) );
			sum4 = s_add( sum4,	s_getA(&buf[i-16] ) );
			}
		sum1 = s_add(sum1, s_add(sum2, s_add(sum3, sum4)) );
		}
		// Tail elements
	
	tmp1 = s_fill( 0 );
	for(	; i > 0;  i -= 4 )
		{
		sum1 = s_add(sum1, tmp1);
		tmp1 = _mm_loadu_ps( (float *) &buf[i-4] );
		}
//	mask = i & 3;	// 0, -1, -2, -3
	tmp1 = s_and( tmp1, SSE_MASK( 4 + i ));
	sum1 = s_add( sum1, tmp1);

	s_put1(	&f_total, s_addp2s(sum1) );
	}

// maxp2f -	
//	function: Given	a 4	packed float input,	return the max among them
//	concept:  Ben Herndon
//
static __forceinline float s_maxp2f( VEC max1 )
	{
	float	tMax;
	VEC		maxReg = max1;
	maxReg = s_copyH2L(	maxReg,	max1 );
	maxReg = s_max(	maxReg,	max1 );		// [0][1] vs [3][4]
	max1 = s_shuffle( maxReg, maxReg, 1 );
	maxReg = s_max1( maxReg, max1 );	// [1] vs [0]

	s_put1(	&tMax, maxReg );
	return ( tMax );
	}

static __forceinline VEC s_maxp2s( VEC max1 )
    {
    VEC     maxReg = max1;
    maxReg  = s_copyH2L(    maxReg, max1 );
    maxReg  = s_max(    maxReg, max1 );     // [0][1] vs [3][4]
    max1    = s_shuffle( maxReg, maxReg, 1 );
    maxReg  = s_max1( maxReg, max1 );       // [1] vs [0]
    return ( maxReg );
    }

#define ahead (6)

tableSum2_ptt( sse_tableSum2 )
	{}
tableSum3_ptt( sse_tableSum3 )
	{return 0;}
tableSum4_ptt( sse_tableSum4 )
	{return 0;}
tableSum5_ptt( sse_tableSum5 )
	{return 0;}

// sse_tableSum2
//
// The function computes the mean of values from two tables
// Since each adjacent float is added to the next simd works well.
// To keep multiple FPU or SSE units in any processor busy we try to
// avoid dependancy chains  ( x += a, x += b, x += c) where each 
// computation must wait til previous sum is computed.
// having two maximums running from two colums of numbers
// avoids a single max register from being the bottleneck
/*
 Typical calling length values taken from a WU run:  LENGTH[ times used ]
 16913[  105] -- 8456[  225] -- 4228[  465] -- 2114[  945] -- 1057[ 5715] 
 --  529[19125] --  264[68,985] --  132[260,865] --   66[1,043,970] --   33[4,118,610]
 --   17[15,480,990] -- 
 Which lengths cause the following di length value in sum2 and sum3
--- Length = 132
pulse tbl3: (di:  44) [ 132]=    0,   44,   88
pulse tbl2: (di:  22) [ 198]=  132,  154,
pulse tbl2: (di:  11) [ 231]=  198,  209,
pulse tbl2: (di:   5) [ 247]=  231,  237,

--- Length = 66
pulse tbl3: (di:  22) [  66]=    0,   22,   44
pulse tbl2: (di:  11) [  99]=   66,   77,
pulse tbl2: (di:   5) [ 115]=   99,  105,

--- Length = 33
pulse tbl3: (di:  11) [  33]=    0,   11,   22
pulse tbl2: (di:   5) [  49]=   33,   39,

--- Length = 17
pulse tbl3: (di:   5) [  17]=    0,    6,   11
*/
sum_func_ptt( sse_sum2_8 )
    {
    VEC     sum1, sum2, tmp1, tmp2;
	int	i=0, length =  P->di;
	float *ptr1 = ss[1]+P->offset;
	float *ptr2 = ss[1]+P->tmp0;
	float *sums = P->dest;

	int mask1 = sse_mask1[ length-i];
	int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	sum1 = s_and(sum1, SSE_MASK(mask1));sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    s_putU( &sums[0], sum1 );           s_putU( &sums[4], sum2 );

	return s_maxp2f( s_max( sum1, sum2 ) );
	}

sum_func_ptt( sse_sum2_12 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i=0, length =  P->di;
	float *ptr1 = ss[1]+P->offset;
	float *ptr2 = ss[1]+P->tmp0;
	float *sums = P->dest;

	int mask1 = sse_mask1[ length-8];
    s_getU(sum1, &ptr1[0] );        s_getU(sum2, &ptr1[4] );
    s_getU(max1, &ptr2[0] );        s_getU(max2, &ptr2[4] );
    max1 = s_add( max1, tmp1 );     max2 = s_add( max2, tmp2 );
    s_putU( &sums[0], max1 );       s_putU( &sums[4], max2 );

    s_getU(sum1, &ptr1[8] );
    s_getU(tmp1, &ptr2[8] );
    sum1 = s_add( sum1, tmp1 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );
    max1 = s_max( max1, sum1 ); 
	s_putU( &sums[8], sum1 );

	return s_maxp2f( s_max( max1, max2 ) );
	}

sum_func_ptt( sse_sum2 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i, length =  P->di;
	float *ptr1 = ss[1]+P->offset;
	float *ptr2 = ss[1]+P->tmp0;
	float *sums = P->dest;

//  divisor = s_fill( 1 / 2.0 );
    max1 = max2 = ZERO;

    const int   stride = 8;

    for (i = 0; i < length-(stride - 1); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );         max2 = s_max( max2, sum2 );
        s_putU( &sums[i + 0], sum1 );       s_putU( &sums[i +4], sum2 );
        }

    int mask1 = sse_mask1[ length-i];
    int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );	sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    max1 = s_max( max1, sum1 );         max2 = s_max( max2, sum2 );
    s_putU( &sums[i + 0], sum1 );       s_putU( &sums[i +4], sum2 );

	return s_maxp2f( s_max( max1, max2 ) );
    }

//
// sse_tableSum3
//
sum_func_ptt( sse_sum3_8 )
    {
    VEC     sum1, sum2, tmp1, tmp2;
	int	i=0, length =  P->di;
	float *ptr1 = ss[1]+P->offset;
	float *ptr2 = ss[1]+P->tmp0;
	float *ptr3 = ss[0]+P->tmp1;
	float *sums = P->dest;

	int mask1 = sse_mask1[ length-i];
	int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	sum1 = s_and(sum1, SSE_MASK(mask1));sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    s_putU( &sums[0], sum1 );           s_putU( &sums[4], sum2 );

	return s_maxp2f( s_max( sum1, sum2 ) );
	}

sum_func_ptt( sse_sum3_12 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i=0, length =  P->di;
	float *ptr1 = ss[1]+P->offset;
	float *ptr2 = ss[1]+P->tmp0;
	float *ptr3 = ss[0]+P->tmp1;
	float *sums = P->dest;

	int mask1 = sse_mask1[ length-8];
    s_getU(sum1, &ptr1[0] );        s_getU(sum2, &ptr1[4] );
    s_getU(max1, &ptr2[0] );        s_getU(max2, &ptr2[4] );
    max1 = s_add( max1, tmp1 );     max2 = s_add( max2, tmp2 );
	s_getU(tmp1, &ptr3[0] );		s_getU(tmp2, &ptr3[4] );
    max1 = s_add( max1, tmp1 );     max2 = s_add( max2, tmp2 );
    s_putU( &sums[0], max1 );       s_putU( &sums[4], max2 );

    s_getU(sum1, &ptr1[8] );
    s_getU(tmp1, &ptr2[8] );
    sum1 = s_add( sum1, tmp1 );
    s_getU(tmp1, &ptr3[8] );
    sum1 = s_add( sum1, tmp1 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );
    max1 = s_max( max1, sum1 ); 
	s_putU( &sums[8], sum1 );

	return s_maxp2f( s_max( max1, max2 ) );
	}

sum_func_ptt( sse_sum3 )
	{
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i, length =  P->di;
	float *ptr1 = ss[0];
	float *ptr2 = ss[0]+P->tmp0;
	float *ptr3 = ss[0]+P->tmp1;
	float *sums = P->dest;

//	divisor	= s_fill( 1	/ 3.0 );
	max1 = max2	= ZERO;

	const int	stride = 8;
	for	(i = 0;	i <	length-(stride - 1); i += stride )
		{
		//	SSE	Pipeline #1						 SSE Pipeline #2
		//
		s_getU(sum1, &ptr1[i + 0] );		s_getU(sum2, &ptr1[i + 4] );
		s_getU(tmp1, &ptr2[i + 0] );		s_getU(tmp2, &ptr2[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
//		sum1 = s_mult( sum1, divisor );		sum2 = s_mult( sum2, divisor );
		max1 = s_max( max1,	sum1 );			max2 = s_max( max2,	sum2 );
		s_putU(	&sums[i	+ 0], sum1 );		s_putU(	&sums[i	+4], sum2 );
		}

    int mask1 = sse_mask1[ length-i];
    int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );	sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    max1 = s_max( max1, sum1 );         max2 = s_max( max2, sum2 );
    s_putU( &sums[i + 0], sum1 );       s_putU( &sums[i +4], sum2 );

	return s_maxp2f( s_max( max1, max2 ) );
	}

//
// sse_tableSum4
//
sum_func_ptt( sse_sum4 )
	{
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i, length =  P->di;
	float *ptr1 = ss[0];
	float *ptr2 = ss[0]+P->tmp0;
	float *ptr3 = ss[0]+P->tmp1;
	float *ptr4 = ss[0]+P->tmp2;
	float *sums = P->dest;

//	divisor	= s_fill( 1	/ 4.0 );
	max1 = max2	= ZERO;

	const int	stride = 8;
	for	(i = 0;	i <	length-(stride - 1); i += stride )
		{
		//	SSE	Pipeline #1						 SSE Pipeline #2
		//
		s_getU(sum1, &ptr1[i + 0] );		s_getU(sum2, &ptr1[i + 4] );
		s_getU(tmp1, &ptr2[i + 0] );		s_getU(tmp2, &ptr2[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr4[i + 0] );		s_getU(tmp2, &ptr4[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
//		sum1 = s_mult( sum1, divisor );		sum2 = s_mult( sum2, divisor );
		max1 = s_max( max1,	sum1 );			max2 = s_max( max2,	sum2 );
		s_putU(	&sums[i	+ 0], sum1 );		s_putU(	&sums[i	+4], sum2 );
		}

    int mask1 = sse_mask1[ length-i];
    int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	s_getU(tmp1, &ptr4[i + 0] );		s_getU(tmp2, &ptr4[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );	sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    max1 = s_max( max1, sum1 );         max2 = s_max( max2, sum2 );
    s_putU( &sums[i + 0], sum1 );       s_putU( &sums[i +4], sum2 );

	return s_maxp2f( s_max( max1, max2 ) );
	}

//
// sse_tableSum5
//
sum_func_ptt( sse_sum5 )
	{
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
	int	i, length =  P->di;
	float *ptr1 = ss[0];
	float *ptr2 = ss[0]+P->tmp0;
	float *ptr3 = ss[0]+P->tmp1;
	float *ptr4 = ss[0]+P->tmp2;
	float *ptr5 = ss[0]+P->tmp3;
	float *sums = P->dest;

//	divisor	= s_fill( 1	/ 5.0 );
	max1 = max2	= ZERO;

	const int	stride = 8;
	for	(i = 0;	i <	length-(stride - 1); i += stride )
		{
		//	SSE	Pipeline #1						 SSE Pipeline #2
		//
		s_getU(sum1, &ptr1[i + 0] );		s_getU(sum2, &ptr1[i + 4] );
		s_getU(tmp1, &ptr2[i + 0] );		s_getU(tmp2, &ptr2[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr4[i + 0] );		s_getU(tmp2, &ptr4[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
		s_getU(tmp1, &ptr5[i + 0] );		s_getU(tmp2, &ptr5[i + 4] );
		sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
//		sum1 = s_mult( sum1, divisor );		sum2 = s_mult( sum2, divisor );
		max1 = s_max( max1,	sum1 );			max2 = s_max( max2,	sum2 );
		s_putU(	&sums[i	+ 0], sum1 );		s_putU(	&sums[i	+4], sum2 );
		}

    int mask1 = sse_mask1[ length-i];
    int mask2 = sse_mask2[ length-i];
    s_getU(sum1, &ptr1[i + 0] );        s_getU(sum2, &ptr1[i + 4] );
    s_getU(tmp1, &ptr2[i + 0] );        s_getU(tmp2, &ptr2[i + 4] );
    sum1 = s_add( sum1, tmp1 );         sum2 = s_add( sum2, tmp2 );
	s_getU(tmp1, &ptr3[i + 0] );		s_getU(tmp2, &ptr3[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	s_getU(tmp1, &ptr4[i + 0] );		s_getU(tmp2, &ptr4[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	s_getU(tmp1, &ptr5[i + 0] );		s_getU(tmp2, &ptr5[i + 4] );
	sum1 = s_add( sum1,	tmp1 );			sum2 = s_add( sum2,	tmp2 );
	sum1 = s_and(sum1, SSE_MASK( mask1 ) );	sum2 = s_and(sum2, SSE_MASK( mask2 ) );
    max1 = s_max( max1, sum1 );         max2 = s_max( max2, sum2 );
    s_putU( &sums[i + 0], sum1 );       s_putU( &sums[i +4], sum2 );

	return s_maxp2f( s_max( max1, max2 ) );
	}


// =============================================================================
//
// sse1_ChirpData_ak
// version by: Alex Kan
//   http://tbp.berkeley.edu/~alexkan/seti/
//
// refactored for SSE by Ben Herndon
// modified by Joe Segur:
//   Added switch to 64 bit precision on Win32 to allow fast rounding
//   Unrolled loop once (to stride 8) to better amortize precision switching

__inline vFloat vec_recip1(vFloat v) 
	{
	// obtain estimate
	vFloat estimate = _mm_rcp_ps( v );
	// one round of Newton-Raphson
	return _mm_add_ps(_mm_mul_ps(_mm_sub_ps(ONE, _mm_mul_ps(estimate, v)), estimate), estimate);
	}  

ChirpData_ptt( sse1_ChirpData_ak )
	{

	if (chirp_rate_ind == 0) {
		memcpy(cx_ChirpDataArray, cx_DataArray,	(int)ul_NumDataPoints * sizeof(sah_complex)	);
		return 0;
		}

	double srate    = chirp_rate * 0.5 / (sample_rate * sample_rate);
	double roundVal = (srate >= 0.0) ? FPUROUND : -FPUROUND;
	int i, vEnd;
	unsigned short fpucw1, fpucw2;

	// main vectorised loop
	vEnd = ul_NumDataPoints - (ul_NumDataPoints & 7);
	for (i = 0; i < vEnd; i += 8) {
		const float *data = (const float *) (cx_DataArray + i);
		float *chirped = (float *) (cx_ChirpDataArray + i);
		double angles[8], dtemp;
		ALIGN_SIMD float  v_angle[4], v_angle2[4];
		
		__m128 d1, d2;
		__m128 cd1, cd2, cd3;
		__m128 td1, td2;
		__m128 x;
		__m128 y;
		__m128 s;
		__m128 c;
		__m128 m;

		// load the signal to be chirped
		s_fetch_temp( data+32 );
		d1 = _mm_load_ps(data);
		d2 = _mm_load_ps(data+4);

		dtemp = double( i );
		angles[0] = dtemp+0.0;
		angles[1] = dtemp+1.0;
		angles[2] = dtemp+2.0;
		angles[3] = dtemp+3.0;
		angles[4] = dtemp+4.0;
		angles[5] = dtemp+5.0;
		angles[6] = dtemp+6.0;
		angles[7] = dtemp+7.0;

		// calculate the input angle
		angles[0] *= angles[0] * srate;	// angle^2 * rate
		angles[1] *= angles[1] * srate;
		angles[2] *= angles[2] * srate;
		angles[3] *= angles[3] * srate;
		angles[4] *= angles[4] * srate;
		angles[5] *= angles[5] * srate;
		angles[6] *= angles[6] * srate;
		angles[7] *= angles[7] * srate;

		// reduce the angle to the range (-0.5, 0.5)
		// convert doubles into singles
		// This attemps to get  "angles - floor(angles)"
#if defined(_WIN32)
	#if defined(_MSC_VER)
		__asm fnstcw fpucw1;
	#elif defined(__LINUX__)
		__asm__ __volatile__ ("fnstcw %0" : "=m" (fpucw1));
	#endif
		fpucw2 = fpucw1 | 0x300;
	#if defined(_MSC_VER)
		__asm
			{
			fldcw fpucw2		// set extended precision
			fld roundVal
			fld angles[0 * 8]	// get angle double
			fadd st(0), st(1)	// + roundVal
			fsub st(0), st(1)	// - roundVal
			fsubr angles[0 * 8]	// angle - round(angle)
			fstp v_angle[0 * 4]	// store as float
			fld angles[1 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[1 * 8]
			fstp v_angle[1 * 4]
			fld angles[2 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[2 * 8]
			fstp v_angle[2 * 4]
			fld angles[3 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[3 * 8]
			fstp v_angle[3 * 4]
			fld angles[4 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[4 * 8]
			fstp v_angle2[0 * 4]
			fld angles[5 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[5 * 8]
			fstp v_angle2[1 * 4]
			fld angles[6 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[6 * 8]
			fstp v_angle2[2 * 4]
			fld angles[7 * 8]
			fadd st(0), st(1)
			fsub st(0), st(1)
			fsubr angles[7 * 8]
			fstp v_angle2[3 * 4]
			fstp st(0)			// pop roundVal
			fldcw fpucw1		// restore normal precision
			}
#elif defined(__GNUC__)
		__asm__ __volatile__ ("fldcw %0" : : "m" (fpucw2));
		v_angle[0] = angles[0] - ((angles[0] + roundVal) - roundVal);
		v_angle[1] = angles[1] - ((angles[1] + roundVal) - roundVal);
		v_angle[2] = angles[2] - ((angles[2] + roundVal) - roundVal);
		v_angle[3] = angles[3] - ((angles[3] + roundVal) - roundVal);
		v_angle2[0] = angles[4] - ((angles[4] + roundVal) - roundVal);
		v_angle2[1] = angles[5] - ((angles[5] + roundVal) - roundVal);
		v_angle2[2] = angles[6] - ((angles[6] + roundVal) - roundVal);
		v_angle2[3] = angles[7] - ((angles[7] + roundVal) - roundVal);
		__asm__ __volatile__ ("fldcw %0" : : "m" (fpucw1));
	#endif
#elif defined(__LINUX__)
		v_angle[0] = angles[0] - ((angles[0] + roundVal) - roundVal);
		v_angle[1] = angles[1] - ((angles[1] + roundVal) - roundVal);
		v_angle[2] = angles[2] - ((angles[2] + roundVal) - roundVal);
		v_angle[3] = angles[3] - ((angles[3] + roundVal) - roundVal);
		v_angle2[0] = angles[4] - ((angles[4] + roundVal) - roundVal);
		v_angle2[1] = angles[5] - ((angles[5] + roundVal) - roundVal);
		v_angle2[2] = angles[6] - ((angles[6] + roundVal) - roundVal);
		v_angle2[3] = angles[7] - ((angles[7] + roundVal) - roundVal);
#else
		v_angle[0] = angles[0] - floor(angles[0]);
		v_angle[1] = angles[1] - floor(angles[1]);
		v_angle[2] = angles[2] - floor(angles[2]);
		v_angle[3] = angles[3] - floor(angles[3]);
		v_angle2[0] = angles[4] - floor(angles[4]);
		v_angle2[1] = angles[5] - floor(angles[5]);
		v_angle2[2] = angles[6] - floor(angles[6]);
		v_angle2[3] = angles[7] - floor(angles[7]);
#endif

		x = _mm_load_ps( v_angle );

		// square to the range [0, 0.25)
		y = _mm_mul_ps(x, x);

		// perform the initial polynomial approximations
		s = _mm_mul_ps(y, SS4);
		c = _mm_mul_ps(y, CC3);						
		s = _mm_add_ps(s, SS3);
		c = _mm_add_ps(c, CC2);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS2);
		c = _mm_add_ps(c, CC1);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS1);
		s = _mm_mul_ps(s, x);
		c = _mm_add_ps(c, ONE);

#if 1
		// perform first angle doubling
		cd1 = _mm_mul_ps(s, c);
		cd2 = _mm_mul_ps(c, c);
		cd3 = _mm_mul_ps(s, s);
		y = _mm_mul_ps(cd1, TWO);
		x = _mm_sub_ps(cd2, cd3);

		// calculate scaling factor to correct the magnitude
		//      m1 = vec_nmsub(y1, y1, vec_nmsub(x1, x1, TWO));
		//      m2 = vec_nmsub(y2, y2, vec_nmsub(x2, x2, TWO));
		cd1 = _mm_mul_ps(x, y);
		cd2 = _mm_mul_ps(x, x);
		cd3 = _mm_mul_ps(y, y);
		m = vec_recip1(_mm_add_ps(cd2, cd3 )); //scaling factor

		// perform second angle doubling
		s = _mm_mul_ps(cd1, TWO);
		c = _mm_sub_ps(cd2, cd3);
#else
		// perform first angle doubling
		x = _mm_sub_ps(_mm_mul_ps(c, c), _mm_mul_ps(s, s));
		y = _mm_mul_ps(_mm_mul_ps(s, c), TWO);

		// calculate scaling factor to correct the magnitude
		//      m1 = vec_nmsub(y1, y1, vec_nmsub(x1, x1, TWO));
		//      m2 = vec_nmsub(y2, y2, vec_nmsub(x2, x2, TWO));
		m = vec_recip1(_mm_add_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y)));

		// perform second angle doubling
		c = _mm_sub_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y));
		s = _mm_mul_ps(_mm_mul_ps(y, x), TWO);
#endif
		// correct the magnitude (final sine / cosine approximations)
		c = _mm_mul_ps(c, m);
		s = _mm_mul_ps(s, m);

		x = s_xor( s_swapRI(d1) , NEG_R);
		y = s_xor( s_swapRI(d2) , NEG_R);
		cd1 = _mm_shuffle_ps(c, c, 0x50);	// 01 01 00 00  AaBb => BBbb => c3c3c4c4
		cd2 = _mm_shuffle_ps(c, c, 0xfa);	// 11 11 10 10  AaBb => AAaa => c1c1c2c2
		td1 = _mm_shuffle_ps(s, s, 0x50);
		td2 = _mm_shuffle_ps(s, s, 0xfa);

		cd1 = _mm_mul_ps(cd1, d1);
		cd2 = _mm_mul_ps(cd2, d2);
		td1 = _mm_mul_ps(td1, x);
		td2 = _mm_mul_ps(td2, y);

		cd1 = _mm_add_ps(cd1, td1);
		cd2 = _mm_add_ps(cd2, td2);

		// store chirped values
		_mm_stream_ps(chirped+0, cd1);
		_mm_stream_ps(chirped+4, cd2);

		d1 = _mm_load_ps(data+8);
		d2 = _mm_load_ps(data+12);

		x = _mm_load_ps( v_angle2 );

		// square to the range [0, 0.25)
		y = _mm_mul_ps(x, x);

		// perform the initial polynomial approximations
		s = _mm_mul_ps(y, SS4);
		c = _mm_mul_ps(y, CC3);						
		s = _mm_add_ps(s, SS3);
		c = _mm_add_ps(c, CC2);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS2);
		c = _mm_add_ps(c, CC1);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS1);
		s = _mm_mul_ps(s, x);
		c = _mm_add_ps(c, ONE);

#if 1
		// perform first angle doubling
		cd1 = _mm_mul_ps(s, c);
		cd2 = _mm_mul_ps(c, c);
		cd3 = _mm_mul_ps(s, s);
		y = _mm_mul_ps(cd1, TWO);
		x = _mm_sub_ps(cd2, cd3);

		// calculate scaling factor to correct the magnitude
		//      m1 = vec_nmsub(y1, y1, vec_nmsub(x1, x1, TWO));
		//      m2 = vec_nmsub(y2, y2, vec_nmsub(x2, x2, TWO));
		cd1 = _mm_mul_ps(x, y);
		cd2 = _mm_mul_ps(x, x);
		cd3 = _mm_mul_ps(y, y);
		m = vec_recip1(_mm_add_ps(cd2, cd3 )); //scaling factor

		// perform second angle doubling
		s = _mm_mul_ps(cd1, TWO);
		c = _mm_sub_ps(cd2, cd3);
#else
		// perform first angle doubling
		x = _mm_sub_ps(_mm_mul_ps(c, c), _mm_mul_ps(s, s));
		y = _mm_mul_ps(_mm_mul_ps(s, c), TWO);

		// calculate scaling factor to correct the magnitude
		//      m1 = vec_nmsub(y1, y1, vec_nmsub(x1, x1, TWO));
		//      m2 = vec_nmsub(y2, y2, vec_nmsub(x2, x2, TWO));
		m = vec_recip1(_mm_add_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y)));

		// perform second angle doubling
		c = _mm_sub_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y));
		s = _mm_mul_ps(_mm_mul_ps(y, x), TWO);
#endif
		// correct the magnitude (final sine / cosine approximations)
		c = _mm_mul_ps(c, m);
		s = _mm_mul_ps(s, m);

		x = s_xor( s_swapRI(d1) , NEG_R);
		y = s_xor( s_swapRI(d2) , NEG_R);
		cd1 = _mm_shuffle_ps(c, c, 0x50);	// 01 01 00 00  AaBb => BBbb => c3c3c4c4
		cd2 = _mm_shuffle_ps(c, c, 0xfa);	// 11 11 10 10  AaBb => AAaa => c1c1c2c2
		td1 = _mm_shuffle_ps(s, s, 0x50);
		td2 = _mm_shuffle_ps(s, s, 0xfa);

		cd1 = _mm_mul_ps(cd1, d1);
		cd2 = _mm_mul_ps(cd2, d2);
		td1 = _mm_mul_ps(td1, x);
		td2 = _mm_mul_ps(td2, y);

		cd1 = _mm_add_ps(cd1, td1);
		cd2 = _mm_add_ps(cd2, td2);

		// store chirped values
		_mm_stream_ps(chirped+8, cd1);
		_mm_stream_ps(chirped+12, cd2);
		}

	if( i < ul_NumDataPoints)
		{  // use original routine to finish up any tailings (max stride-1 elements)
		orig_ChirpData(cx_DataArray+i, cx_ChirpDataArray+i
			, chirp_rate_ind, chirp_rate, ul_NumDataPoints-i, sample_rate);
		}
	ex_count_flops( 12* ul_NumDataPoints );

	return 0;
	}


//
// copy_SSE - Use Non temporal writes, but regular pre-fetches 
//   only available with 3DNow+ and SSE+ - Athlon and Pentium III
//
void  copy_SSE(char *dest, const char *src, int blocks)
	{
	VEC temp0, temp1, temp2, temp3;

	float *f_src = (float*)src;
	float *f_dst = (float*)dest;
	s_fetch( f_src );
	s_fetch( f_src+16 );
	if( ((size_t)src) & 15 ) // misaligned source
		while( blocks-- )
			{
			s_fetch( f_src + 32 );
			s_getU(temp0, &f_src[ 0] );
			s_getU(temp1, &f_src[ 4] );
			s_getU(temp2, &f_src[ 8] );
			s_getU(temp3, &f_src[12] );
			s_putA_nc( f_dst +  0, temp0 );
			s_putA_nc( f_dst +  4, temp1 );
			s_putA_nc( f_dst +  8, temp2 );
			s_putA_nc( f_dst + 12, temp3 );
			f_src += 16, f_dst += 16;
			}
	else // aligned source, use movaps instead of movups
		{
		while( blocks-- )
			{
			s_fetch( f_src + 32 );
			temp0 = s_getA( f_src + 0 );
			temp1 = s_getA( f_src + 4 );
			temp2 = s_getA( f_src + 8 );
			temp3 = s_getA( f_src +12 );
			s_putA_nc( f_dst + 0 , temp0 );
			s_putA_nc( f_dst + 4 , temp1 );
			s_putA_nc( f_dst + 8 , temp2 );
			s_putA_nc( f_dst +12, temp3 );
			f_src += 16, f_dst += 16;
			}
		}
	s_fence_writes();
	}


// Arrays from which folding subroutines will be chosen during execution.
// Other sets can be inserted to replace the defaults.
//

sum_func sse_list3[FOLDTBLEN] = {
//	sse_sum3_8, sse_sum3_8, sse_sum3_8, sse_sum3_8,
//	sse_sum3_8, sse_sum3_8, sse_sum3_8, sse_sum3_8,
//	sse_sum3_12, sse_sum3_12, sse_sum3_12, sse_sum3_12,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3,
	sse_sum3, sse_sum3, sse_sum3, sse_sum3
};
sum_func sse_list4[FOLDTBLEN] = {
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4,
	sse_sum4, sse_sum4, sse_sum4, sse_sum4
};
sum_func sse_list5[FOLDTBLEN] = {
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5,
	sse_sum5, sse_sum5, sse_sum5, sse_sum5
};
sum_func sse_list2[FOLDTBLEN] = {
//	sse_sum2_8, sse_sum2_8, sse_sum2_8, sse_sum2_8,
//	sse_sum2_8, sse_sum2_8, sse_sum2_8, sse_sum2_8,
//	sse_sum2_12, sse_sum2_12, sse_sum2_12, sse_sum2_12,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2
};
sum_func sse_list2_L[FOLDTBLEN] = {
//	sse_sum2_8, sse_sum2_8, sse_sum2_8, sse_sum2_8,
//	sse_sum2_8, sse_sum2_8, sse_sum2_8, sse_sum2_8,
//	sse_sum2_12, sse_sum2_12, sse_sum2_12, sse_sum2_12,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2,
	sse_sum2, sse_sum2, sse_sum2, sse_sum2
};

FoldSet sse_ben_fold = {sse_list3, sse_list4, sse_list5, sse_list2, sse_list2_L, "ben SSE"};



//
// PowerSpectrum functions.
//
// Original author: Ben Herndon
// Adapted by Joe Segur
//
GetPowerSpectrum_ptt( sse_GetTPS_sc8 )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j, bin;
    float *workBuf = (float *)FreqData;
    const int stride = 8;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0, bin = 0;
          j < this_fft_len - (stride - 1) ;
          i += stride * 2, j += stride, bin += bin_len * stride)
        {

        //      SSE Pipeline #1
            //      SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );     
        sum1  = s_mult( sum1, sum1);
            sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);
            sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);
            tmp2  = s_swapRI(sum3);
        s_put1( &PowerSpectrum[bin_off + bin + 0 * bin_len], sum1);
            sum3  = s_add(sum3, tmp2);
        sum2  = s_getA( &workBuf[i+1*4] );
            s_put1( &PowerSpectrum[bin_off + bin + 4 * bin_len], sum3);
        sum2  = s_mult( sum2, sum2);
            sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);
            sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);
            tmp2  = s_swapRI(sum4);
        s_put1( &PowerSpectrum[bin_off + bin + 2 * bin_len], sum2);
            sum4  = s_add(sum4, tmp2);
        sum1  = s_shuffle(sum1, sum2, 0x88);
            s_put1( &PowerSpectrum[bin_off + bin + 6 * bin_len], sum4);
        s_putA(&workBuf[j+0*4], sum1);
            sum3  = s_shuffle(sum3, sum4, 0x88);
        sum1  = s_swapRI(sum1); 
            s_putA(&workBuf[j+1*4], sum3);
        s_put1( &PowerSpectrum[bin_off + bin + 1 * bin_len], sum1);
            sum3  = s_swapRI(sum3);
        sum1  = s_shuffle(sum1, sum1, 0xC6);
            s_put1( &PowerSpectrum[bin_off + bin + 5 * bin_len], sum3);
        s_put1( &PowerSpectrum[bin_off + bin + 3 * bin_len], sum1);
            sum3  = s_shuffle(sum3, sum3, 0xC6);
            s_put1( &PowerSpectrum[bin_off + bin + 7 * bin_len], sum3);
        }
    // Should never be used - FFT length < 8 or not power of 2
    for (  ; j < this_fft_len; j++, bin += bin_len)
        {
        workBuf[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
        PowerSpectrum[bin_off + bin] = workBuf[j];
        }
    }

//
// version with no prefetch
//
PwrSpectrumOnly_ptt( sse_GetPSO_sc16_npr )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j;
    float *workBuf  = (float *)FreqData; // 0, 2, 4... real, 1, 3, 5... imag
    const int stride = 16;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0; j < this_fft_len - (stride - 1) ; i += stride * 2, j += stride)
        {
        //      SSE Pipeline #1
                                                //SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+1*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+0*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+1*4], sum3);
        sum1  = s_getA( &workBuf[i+4*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+6*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+5*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+7*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+2*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+3*4], sum3);
        }
    // This code will only be used for fft length < 16 at ar > 701 (an evil tester,
    // a large splitter error, or revised workunit header values)
    for (  ; j < this_fft_len; j++)
        PowerSpectrum[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
    }
//
// version prefetching for cache linesize = 32
//
PwrSpectrumOnly_ptt( sse_GetPSO_sc16_p32 )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j;
    float *workBuf  = (float *)FreqData; // 0, 2, 4... real, 1, 3, 5... imag
    const int stride = 16;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0; j < this_fft_len - (stride - 1) ; i += stride * 2, j += stride)
        {
        char    *ptr = (char *)&workBuf[i+16*16];
        s_fetch_temp (ptr+0*32);
        s_fetch_temp (ptr+1*32);        // if linesize = 32

        //      SSE Pipeline #1
                                                //SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+1*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+0*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+1*4], sum3);
        s_fetch_temp (ptr+2*32);
        s_fetch_temp (ptr+3*32);

        sum1  = s_getA( &workBuf[i+4*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+6*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+5*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+7*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+2*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+3*4], sum3);
        }
    // This code will only be used for fft length < 16 at ar > 701 (an evil tester,
    // a large splitter error, or revised workunit header values)
    for (  ; j < this_fft_len; j++)
        PowerSpectrum[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
    }
//
// version prefetching for cache linesize=64
//
PwrSpectrumOnly_ptt( sse_GetPSO_sc16_p64 )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j;
    float *workBuf  = (float *)FreqData; // 0, 2, 4... real, 1, 3, 5... imag
    const int stride = 16;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0; j < this_fft_len - (stride - 1) ; i += stride * 2, j += stride)
        {
        char    *ptr = (char *)&workBuf[i+16*16];
        s_fetch_temp (ptr);

        //      SSE Pipeline #1
                                                //SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+1*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+0*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+1*4], sum3);
        s_fetch_temp (ptr+64);

        sum1  = s_getA( &workBuf[i+4*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+6*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+5*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+7*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+2*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+3*4], sum3);
        }
    // This code will only be used for fft length < 16 at ar > 701 (an evil tester,
    // a large splitter error, or revised workunit header values)
    for (  ; j < this_fft_len; j++)
        PowerSpectrum[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
    }
//
// version prefetching for cache linesize=128
//
PwrSpectrumOnly_ptt( sse_GetPSO_sc16_p128 )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j;
    float *workBuf  = (float *)FreqData; // 0, 2, 4... real, 1, 3, 5... imag
    const int stride = 16;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0; j < this_fft_len - (stride - 1) ; i += stride * 2, j += stride)
        {
        char    *ptr = (char *)&workBuf[i+16*16];
        s_fetch_temp (ptr);

        //      SSE Pipeline #1
                                                //SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+1*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+0*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+1*4], sum3);
        sum1  = s_getA( &workBuf[i+4*4] );
        sum1  = s_mult( sum1, sum1);            sum3  = s_getA( &workBuf[i+6*4] );
        tmp1  = s_swapRI(sum1);                 sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);              tmp2  = s_swapRI(sum3);
        sum2  = s_getA( &workBuf[i+5*4] );     sum3  = s_add(sum3, tmp2);
        sum2  = s_mult( sum2, sum2);            sum4  = s_getA( &workBuf[i+7*4] );
        tmp1  = s_swapRI(sum2);                 sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);              tmp2  = s_swapRI(sum4);
        sum1  = s_shuffle(sum1, sum2, 0x88);    sum4  = s_add(sum4, tmp2);
        s_putA(&PowerSpectrum[j+2*4], sum1);    sum3  = s_shuffle(sum3, sum4, 0x88);
                                                s_putA(&PowerSpectrum[j+3*4], sum3);
        }
    // This code will only be used for fft length < 16 at ar > 701 (an evil tester,
    // a large splitter error, or revised workunit header values)
    for (  ; j < this_fft_len; j++)
        PowerSpectrum[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
    }

// JWS Transpose routines, modified to do a strip of exactly 4 lines.
//
// Common sub 4x4 transpose in the SSE registers, non-temporal writes.
// An sfence is needed after using this sub to ensure global visibilty of the writes.
//
inline void v_vsubTranspose4ntw(float *in, float *out, int xline, int yline)
    {
    __m128 tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;

    tmp1 = _mm_load_ps(in);                 // a3 a2 a1 a0
    tmp2 = _mm_load_ps(in+xline);           // b3 b2 b1 b0
    tmp3 = _mm_load_ps(in+2*xline);         // c3 c2 c1 c0
    tmp4 = _mm_load_ps(in+3*xline);         // d3 d2 d1 d0
    tmp5 = tmp1;                            // a3 a2 a1 a0
    tmp6 = tmp3;                            // c3 c2 c1 c0
    tmp1 = _mm_unpacklo_ps(tmp1,tmp2);      // b1 a1 b0 a0
    tmp3 = _mm_unpacklo_ps(tmp3,tmp4);      // d1 c1 d0 c0
    tmp5 = _mm_unpackhi_ps(tmp5,tmp2);      // b3 a3 b2 a2
    tmp6 = _mm_unpackhi_ps(tmp6,tmp4);      // d3 c3 d2 c2
    tmp2 = tmp1;                            // b1 a1 b0 a0
    tmp1 = _mm_movelh_ps(tmp1,tmp3);        // d0 c0 b0 a0
    tmp3 = _mm_movehl_ps(tmp3,tmp2);        // d1 c1 b1 a1
    _mm_stream_ps(out+0*yline, tmp1);
    tmp2 = tmp5;                            // b3 a3 b2 a2
    _mm_stream_ps(out+1*yline, tmp3);
    tmp5 = _mm_movelh_ps(tmp5,tmp6);        // d2 c2 b2 a2
    tmp6 = _mm_movehl_ps(tmp6,tmp2);        // d3 c3 b3 a3
    _mm_stream_ps(out+2*yline, tmp5);
    _mm_stream_ps(out+3*yline, tmp6);
    }

// x = input line length (FFT length)
// y = output column length (Data points in WU / FFT length)
// r = beginning line number (FFT chunk)

TransposeStrip_ptt( sse_vTranspose4ntw )
    {
    int i, j;
    ALIGNED_YES( in );
    ALIGNED_YES( out );
    for (i = 0; i < x - 3; i += 4)
        {
        v_vsubTranspose4ntw(in + i, out + y * i + r, x, y);
        }
    for (     ; i < x ; i++) // tail should never
        {
        for (j = 0; j < 4; j++) out[i * y + r + j] = in[j * x + i];
        }
    _mm_sfence();
    }

TransposeStrip_ptt( sse_vpfTranspose8x4ntw )
    {
    int i, j;
    ALIGNED_YES( in );
    ALIGNED_YES( out );
    for (i = 0; i < x - 7; i += 8)
        {
        float *tin = in + i;
        _mm_prefetch((char *)(tin + 0 * x + 8), _MM_HINT_T0);
        v_vsubTranspose4ntw(tin, out + y * i + r, x, y);
        _mm_prefetch((char *)(tin + 1 * x + 8), _MM_HINT_T0);
        v_vsubTranspose4ntw(tin + 4, out + y * (i + 4) + r, x, y);
        _mm_prefetch((char *)(tin + 2 * x + 8), _MM_HINT_T0);
        _mm_prefetch((char *)(tin + 3 * x + 8), _MM_HINT_T0);
        }
    for (     ; i < x ; i++) // tail should never
        {
        for (j = 0; j < 4; j++) out[i * y + r + j] = in[j * x + i];
        }
    _mm_sfence();
    }
