/* name: opt_unopt.cpp 
 *
 * Purpose: functions as they would be without any programmer optimizations 
 * Originator: Ben Herndon 

 * These are original functions or code sequences from the unmodified 
 *  seti@home source code.
 *
 * TO OPTMIZE - make copies of these functions and put them in other files
 *   and perform the optimization there...leave these functions as they are.
 *
 * This file should not be compiled with any compiler SIMD enabling options
 *  as these functions serve as a baseline for optimzation improvement 
 *  timing tests
 *
 */

#include "stdafx.h"

#include <mathimf.h>

// ORIGINAL c++ coded functions called inside find_pulse
#pragma optimize( "p", on )


// std_rebin
lower_rebin_ptt( std_rebin )
    {
    int last;
    last = length / res;
    for ( int j = 0; j < last; j++ )
        {
        float   sum = 0;
        for ( int i = 0; i < res; i++ )
            {
            sum += fp_pulse[j * res + i];
            }

        buf[j] = sum / res;
        }
    }

// std_rms_deviation
rms_dev_ptt( std_rms_deviation )
    {
    double  rms;
    for ( int i = 0; i < length; i++ )
        {
        float   num;
        num = buf[i] - avg;

        rms += num * num;
        }

    return ( rms );
    }

// std_f_sum
f_sum_ptt( std_f_sum )
    {
    float   sum = 0;
    for ( int i = length; i; i-- )
        {
        sum += buf[i - 1];
        }

	f_total = sum;
    }

// std_tableSum2
tableSum2_ptt( std_tableSum2 )
    {
    float tmp_max = 0;

    for ( int i = 0; i < length; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ( ptr1[i] + ptr2[i] ) / 2;
        sums[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }
	f_max = tmp_max;
    }

// std_tableSum3
tableSum3_ptt( std_tableSum3 )
    {
    float   tmp_max;
    tmp_max = 0;

    for ( int i = 0; i < length; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ( ptr1[i] + ptr2[i] + ptr3[i] ) / 3;
        sums[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }

    return ( tmp_max );
    }

// std_tableSum4
tableSum4_ptt( std_tableSum4 )
    {
    float   tmp_max;
    tmp_max = 0;

    for ( int i = 0; i < length; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ( ptr1[i] + ptr2[i] + ptr3[i] + ptr4[i] ) / 4;
        sums[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }

    return ( tmp_max );
    }

// std_tableSum5
tableSum5_ptt( std_tableSum5 )
    {
    float   tmp_max;
    tmp_max = 0;

    for ( int i = 0; i < length; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ( ptr1[i] + ptr2[i] + ptr3[i] + ptr4[i] + ptr5[i] ) / 5;
        sums[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }

    return ( tmp_max );
    }



#pragma optimize( "p", off )

// orig_ChirpData 
// xx-xx-03 ERICK -  Created function
//
ChirpData_ptt( orig_ChirpData )
    {
	if ( chirp_rate_ind == 0 )
        {
        memcpy( cx_ChirpDataArray, cx_DataArray,  int( ul_NumDataPoints * sizeof(sah_complex)) );  
        return 0 ;
        }

	ALIGNED_YES( cx_ChirpDataArray );
	ALIGNED_YES( cx_DataArray );

	for	(int i	= 0; i < ul_NumDataPoints; i++ )
		{
		float	real,	imag;
		double	dd,	cc;
		double	time;
		double	ang;

		// since ang is	getting	moded by 2pi, we calculate "ang	mod
		// 2pi"	before the call	to sincos()	inorder	to reduce
		// roundoff	error. (Bug	submitted by Tetsuji "Maverick"	Rai)
		time = static_cast<	double >( i	) /	sample_rate;
		ang	= 0.5 *	chirp_rate * time *	time;

		ang	-= floor( ang );
		ang	*= M_PI	* 2;
		sincos(	ang, &dd, &cc );

		// Sometimes chirping is done in place.	We don't want to
		// overwrite data prematurely.
		real = cx_DataArray[i][0] *	cc -	cx_DataArray[i][1] * dd;
		imag = cx_DataArray[i][0] *	dd +	cx_DataArray[i][1] * cc;
		cx_ChirpDataArray[i][0]	= real;
		cx_ChirpDataArray[i][1]	= imag;
		}

	count_flops( 12 * ul_NumDataPoints );
    return ( 0 );
    }
 
