/* optimize.hpp:
 * prototypes and functions that merit optimization
 *
 * Originator: Ben Herndon
 * Definitions of function prototypes and function pointers	for functions
 *  that merit optimization within set@home worker app
 */
#include "../client/config.h"
#pragma once

#ifndef	optimize_H
	#define	optimize_H

#undef M_PI
#define	M_PI	3.14159265358979323846264338327950

const unsigned int _b1111 = 0xFFFFFFFF;
const unsigned int _b0000 = 0;

#define BEN_SPECTRUM 1

extern bool	optimize_init( void	);

extern void  ex_count_flops(int the_num);

extern bool show_benchmark;
extern int L1_cache;       // Let rest of code know cache sizes
extern int L2_cache;
extern int L1_line;
extern int worker_cnt;  // How many cores/cpus share the Cache
extern double force_use;

/* =============================================================================
 function prototypes / macros for functions used inside "pulsefind.cpp"
 =============================================================================== */
typedef	float const ( *c_fp );
typedef	float sah_complex[2];

// parameter macros
	#define	lower_rebin_ptt( func_ )	 \
		void  func_( float *buf, c_fp fp_pulse,	int	length,	int	res	)
	#define	rms_dev_ptt( func_ )	\
		float  func_( c_fp buf, int length, float avg )

	#define	f_sum_ptt( func_ )		\
		void  func_( float &f_total, c_fp buf, int length )
	#define	tableSum2_ptt( func_ )	\
		void  func_( float &f_max, int length, float *sums, c_fp ptr1, c_fp ptr2 )
	#define	tableSum3_ptt( func_ ) \
	float  func_	(				   \
			int	length,			   \
			float *sums,		   \
			c_fp ptr1,			   \
			c_fp ptr2,			   \
			c_fp ptr3 )
	#define	tableSum4_ptt( func_ ) \
	float  func_	(				   \
			int	length,			   \
			float *sums,		   \
			c_fp ptr1,			   \
			c_fp ptr2,			   \
			c_fp ptr3,			   \
			c_fp ptr4 )
	#define	tableSum5_ptt( func_ ) \
	float  func_	(				   \
			int	length,			   \
			float *sums,		   \
			c_fp ptr1,			   \
			c_fp ptr2,			   \
			c_fp ptr3,			   \
			c_fp ptr4,			   \
			c_fp ptr5 )

typedef					lower_rebin_ptt( lower_rebin_func );
typedef					f_sum_ptt( f_sum_func );
typedef					rms_dev_ptt( rms_dev_func );
typedef					tableSum2_ptt( tableSum2_func );
typedef					tableSum3_ptt( tableSum3_func );
typedef					tableSum4_ptt( tableSum4_func );
typedef					tableSum5_ptt( tableSum5_func );

// function	pointers - used	to replace original	function calls
extern lower_rebin_func	*lower_rebin;
extern f_sum_func		*f_sum;
extern rms_dev_func		*rms_deviation;
extern tableSum2_func	*tableSum2;
extern tableSum3_func	*tableSum3;
extern tableSum4_func	*tableSum4;
extern tableSum5_func	*tableSum5;

// "standard" function names - renamed original	function names
extern lower_rebin_func	std_rebin;
extern f_sum_func		std_f_sum;
extern rms_dev_func		std_rms_deviation;
extern tableSum2_func	std_tableSum2;
extern tableSum3_func	std_tableSum3;
extern tableSum4_func	std_tableSum4;
extern tableSum5_func	std_tableSum5;

extern f_sum_func		opt_f_sum;
extern tableSum2_func	opt_tableSum2;
extern tableSum3_func	opt_tableSum3;
extern tableSum4_func	opt_tableSum4;
extern tableSum5_func	opt_tableSum5;

extern f_sum_func       sse_f_sum;
// extern rms_dev_func  sse_rms_deviation;
extern tableSum2_func   sse_tableSum2;
extern tableSum3_func   sse_tableSum3;
extern tableSum4_func   sse_tableSum4;
extern tableSum5_func   sse_tableSum5;

// =============================================================================
//	  function	prototypes macros for functions	used inside	"gaussfit.cpp"
// =============================================================================

#define	f_GetPeak_ptt( func_ )	 \
	float  func_(		 \
		c_fp	fp_PoT,			 \
		int		ul_TOffset,		 \
		int		ul_HalfSumLength,\
		float	f_MeanPower,	 \
		float	f_PeakScaleFactor, \
		c_fp	f_weight		\
		)

#define	f_GetChiSq_ptt( func_ )	\
	float  func_(				\
        c_fp	fp_PoT,			\
        int     ul_PowerLen,	\
        int     ul_TOffset,		\
        float   f_PeakPower,	\
        float   f_MeanPower,	\
        c_fp	f_weight,		\
		float	rebin,			\
        float   *xsq_null		\
    )

typedef	f_GetPeak_ptt( f_GetPeak_func );
typedef	f_GetChiSq_ptt( f_GetChiSq_func );

// function	pointers - used	to replace original	function calls
extern f_GetPeak_func	*f_GetPeak;
extern f_GetChiSq_func	*f_GetChiSq;

// "standard" function names - renamed original	function names
extern f_GetPeak_func	std_f_GetPeak;
extern f_GetChiSq_func	std_f_GetChiSq;

extern f_GetPeak_func	opt_f_GetPeak;
extern f_GetChiSq_func	opt_f_GetChiSq;

extern f_GetPeak_func   sse_f_GetPeak;

// =============================================================================
//	  function	prototypes macros for "analyzeFuncs.cpp"
// =============================================================================

#define	ChirpData_ptt( func_ )				\
int	func_(										\
		sah_complex const	*cx_DataArray,		\
		sah_complex			*cx_ChirpDataArray,	\
		int					chirp_rate_ind,		\
		double				chirp_rate,			\
		int					ul_NumDataPoints,	\
		double				sample_rate )

#define	TrigArrayInit_ptt( func_ )				\
void func_(										\
		int					len,				\
		double				ChirpStep,			\
		int					InitChirpRateInd,	\
		double				SampleRate )

#define	TrigArrayFree_ptt( func_ )				\
void func_( void )

#define	GetPowerSpectrum_ptt( func_	) \
void func_ ( sah_complex *FreqData,	\
		float *PowerSpectrum,		\
		int this_fft_len,			\
		int bin_off, int bin_len )

#define	PwrSpectrumOnly_ptt( func_	) \
void func_ ( sah_complex *FreqData,	\
		float *PowerSpectrum,		\
		int this_fft_len)

#define	TransposeStrip_ptt( func_ ) \
void func_ ( int x, int y, int r,	\
		float *in, float *out)

typedef							ChirpData_ptt( ChirpData_func );
typedef							TrigArrayInit_ptt( TrigArrayInit_func );
typedef							TrigArrayFree_ptt( TrigArrayFree_func );
typedef							GetPowerSpectrum_ptt( GetPowerSpectrum_func	);
typedef							PwrSpectrumOnly_ptt( PwrSpectrumOnly_func );
typedef							TransposeStrip_ptt( TransposeStrip_func );

// function	pointers - used	to replace original	function calls
extern ChirpData_func			*ChirpData;
extern TrigArrayInit_func		*TrigArrayInit;
extern TrigArrayFree_func		*TrigArrayFree;
extern GetPowerSpectrum_func	*GetPowerSpectrum;
extern PwrSpectrumOnly_func 	*PwrSpectrumOnly;
extern TransposeStrip_func		*TransposeStrip;

// "standard" function names - renamed original	function names
extern ChirpData_func			orig_ChirpData;
extern ChirpData_func			std_v_ChirpData;
extern GetPowerSpectrum_func	std_v_GetPowerSpectrum;
extern PwrSpectrumOnly_func 	orig_v_GetPowerSpectrum;
extern TransposeStrip_func		orig_v_Transpose;
extern TransposeStrip_func		orig_v_Transpose2;
extern TransposeStrip_func		orig_v_Transpose4;

extern ChirpData_func			opt_ChirpData;
extern GetPowerSpectrum_func	opt_v_GetPowerSpectrum;

extern GetPowerSpectrum_func    sse_v_GetPowerSpectrum;
extern GetPowerSpectrum_func	sse_GetTPS_sc8;
extern PwrSpectrumOnly_func 	sse_GetPSO_sc16_npr;
extern PwrSpectrumOnly_func 	sse_GetPSO_sc16_p32;
extern PwrSpectrumOnly_func 	sse_GetPSO_sc16_p64;
extern PwrSpectrumOnly_func 	sse_GetPSO_sc16_p128;
extern TransposeStrip_func		sse_vTranspose4ntw;
extern TransposeStrip_func		sse_vpfTranspose8x4ntw;

extern GetPowerSpectrum_func    sse2_v_GetPowerSpectrum;
extern GetPowerSpectrum_func	sse2_GetTPS_sc8;
extern ChirpData_func			sse2_ChirpData;
extern ChirpData_func			sse2_ChirpData_ak;

// =============================================================================
// macros to help out with different compilers
// =============================================================================


// ALIGNED_YES Macro - tell	compiler buffer	is already SIMD	aligned,
// optimize	appropriately
//

#define SIMD_ALIGN 16

	// Make variable lie on correct boundary
#ifdef __LINUX__
#define ALIGN_SIMD __attribute__ ((aligned(SIMD_ALIGN)))
#else
#define ALIGN_SIMD __declspec(align(SIMD_ALIGN))
#endif

	// Tell compiler "pointer points to aligned data"
#if defined( __INTEL_COMPILER )
	#define	ALIGNED_YES( buffer_ )	__assume_aligned( buffer_, SIMD_ALIGN );
#else
	#define	ALIGNED_YES( buffer_ )
#endif


#endif // optimize_H
